"""
Mocks for the `class`
"""

from ast import (
    AnnAssign,
    Assign,
    Attribute,
    BinOp,
    Call,
    ClassDef,
    Expr,
    FunctionDef,
    Index,
    List,
    Load,
    Mult,
    Name,
    Pass,
    Return,
    Store,
    Sub,
    Subscript,
    Tuple,
    UnaryOp,
    USub,
    arguments,
    keyword,
)
from copy import deepcopy
from functools import partial
from itertools import chain
from operator import add, itemgetter
from textwrap import indent

from cdd.shared.ast_utils import maybe_type_comment, set_arg, set_slice, set_value
from cdd.shared.defaults_utils import extract_default
from cdd.shared.pure_utils import strip_starting, tab
from cdd.tests.mocks.docstrings import docstring_header_str, docstring_reduction_v2_str
from cdd.tests.mocks.methods import (
    function_google_tf_squared_hinge_docstring_str,
    returns_subscript,
)
from cdd.tests.utils_for_tests import remove_args_from_docstring

class_doc_str = tab.join(
    (
        "\n",
        "{header_doc_str}{tab}\n".format(header_doc_str=docstring_header_str, tab=tab),
        ':cvar dataset_name: name of dataset. Defaults to "mnist"\n',
        ':cvar tfds_dir: directory to look for models in. Defaults to "~/tensorflow_datasets"\n',
        ':cvar K: backend engine, e.g., `np` or `tf`. Defaults to "np"\n',
        ":cvar as_numpy: Convert to numpy ndarrays\n",
        ":cvar data_loader_kwargs: pass this as arguments to data_loader function\n",
        ":cvar return_type: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))",
    )
)
class_doc_str_expr = Expr(set_value(class_doc_str))

class_str = '''
class ConfigClass(object):
    """
{header_doc_str}{tab}
    :cvar dataset_name: name of dataset. Defaults to "mnist"
    :cvar tfds_dir: directory to look for models in. Defaults to "~/tensorflow_datasets"
    :cvar K: backend engine, e.g., `np` or `tf`. Defaults to "np"
    :cvar as_numpy: Convert to numpy ndarrays
    :cvar data_loader_kwargs: pass this as arguments to data_loader function
    :cvar return_type: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))"""

    dataset_name: str = "mnist"
    tfds_dir: str = "~/tensorflow_datasets"
    K: Literal["np", "tf"] = "np"
    as_numpy: Optional[bool]
    data_loader_kwargs: Optional[dict]
    return_type: Union[
        Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]
    ] = (
        np.empty(0),
        np.empty(0),
    )
'''.format(
    header_doc_str=indent(docstring_header_str, tab), tab=tab
)

class_nargs_str = '''
class ConfigClass(object):
    """
    {header_doc_str}

    :cvar callbacks: Collection of callables that are run inside the training loop"""

    callbacks: Optional[
        List[
            Literal[
                "BaseLogger",
                "CSVLogger",
                "Callback",
                "CallbackList",
                "EarlyStopping",
                "History",
                "LambdaCallback",
                "LearningRateScheduler",
                "ModelCheckpoint",
                "ProgbarLogger",
                "ReduceLROnPlateau",
                "RemoteMonitor",
                "TensorBoard",
                "TerminateOnNaN",
            ]
        ]
    ] = None
'''.format(
    header_doc_str=indent(docstring_header_str, tab)
)

class_ast = ClassDef(
    bases=[Name("object", Load())],
    body=[
        class_doc_str_expr,
        AnnAssign(
            annotation=Name(
                "str",
                Load(),
            ),
            simple=1,
            target=Name("dataset_name", Store()),
            value=set_value("mnist"),
            expr=None,
            expr_annotation=None,
            expr_target=None,
        ),
        AnnAssign(
            annotation=Name(
                "str",
                Load(),
            ),
            simple=1,
            target=Name("tfds_dir", Store()),
            value=set_value(
                "~/tensorflow_datasets",
            ),
            expr=None,
            expr_annotation=None,
            expr_target=None,
        ),
        AnnAssign(
            annotation=Subscript(
                Name(
                    "Literal",
                    Load(),
                ),
                Index(
                    value=Tuple(
                        elts=list(
                            map(
                                set_value,
                                (
                                    "np",
                                    "tf",
                                ),
                            )
                        ),
                        ctx=Load(),
                        expr=None,
                    )
                ),
                Load(),
            ),
            simple=1,
            target=Name("K", Store()),
            value=set_value("np"),
            expr=None,
            expr_target=None,
            expr_annotation=None,
        ),
        AnnAssign(
            annotation=Subscript(
                Name(
                    "Optional",
                    Load(),
                ),
                Index(value=Name("bool", Load())),
                Load(),
            ),
            simple=1,
            target=Name("as_numpy", Store()),
            value=None,
            expr=None,
            expr_target=None,
            expr_annotation=None,
        ),
        AnnAssign(
            annotation=Subscript(
                Name("Optional", Load()), set_slice(Name("dict", Load())), Load()
            ),
            simple=1,
            target=Name(
                "data_loader_kwargs",
                Store(),
            ),
            value=None,
            expr=None,
            expr_target=None,
            expr_annotation=None,
        ),
        AnnAssign(
            annotation=returns_subscript,
            simple=1,
            target=Name("return_type", Store()),
            value=Tuple(
                ctx=Load(),
                elts=[
                    Call(
                        args=[set_value(0)],
                        func=Attribute(
                            Name("np", Load()),
                            "empty",
                            Load(),
                        ),
                        keywords=[],
                        expr=None,
                        expr_func=None,
                    )
                ]
                * 2,
                expr=None,
            ),
            expr=None,
            expr_target=None,
            expr_annotation=None,
        ),
    ],
    decorator_list=[],
    keywords=[],
    name="ConfigClass",
    expr=None,
    identifier_name=None,
)

class_ast_no_default_doc = deepcopy(class_ast)
class_ast_no_default_doc.body[0] = Expr(
    set_value(
        "\n".join(
            map(
                itemgetter(0),
                map(
                    partial(extract_default, emit_default_doc=False),
                    class_doc_str.splitlines(),
                ),
            )
        )
    )
)

class_ast_with_none = deepcopy(class_ast)
assert (
    isinstance(class_ast_with_none.body[4], AnnAssign)
    and class_ast_with_none.body[4].target.id == "as_numpy"
)
class_ast_with_none.body[4].value = set_value(
    None
)  # E.g., because argparse has a default set to `None`

class_nargs_ast = ClassDef(
    bases=[Name("object", Load())],
    body=[
        Expr(
            set_value(
                "\n{tab}{header_doc_str}{tab}\n{tab}"
                ":cvar callbacks: Collection of callables that are run inside the training loop".format(
                    tab=tab, header_doc_str=docstring_header_str
                ),
            )
        ),
        AnnAssign(
            annotation=Subscript(
                Name("Optional", Load()),
                Index(
                    value=Subscript(
                        Name("List", Load()),
                        Index(
                            value=Subscript(
                                Name("Literal", Load()),
                                Index(
                                    value=Tuple(
                                        ctx=Load(),
                                        elts=list(
                                            map(
                                                set_value,
                                                (
                                                    "BaseLogger",
                                                    "CSVLogger",
                                                    "Callback",
                                                    "CallbackList",
                                                    "EarlyStopping",
                                                    "History",
                                                    "LambdaCallback",
                                                    "LearningRateScheduler",
                                                    "ModelCheckpoint",
                                                    "ProgbarLogger",
                                                    "ReduceLROnPlateau",
                                                    "RemoteMonitor",
                                                    "TensorBoard",
                                                    "TerminateOnNaN",
                                                ),
                                            )
                                        ),
                                        expr=None,
                                    )
                                ),
                                Load(),
                            )
                        ),
                        Load(),
                    )
                ),
                Load(),
            ),
            simple=1,
            target=Name("callbacks", Store()),
            value=None,
            expr=None,
            expr_annotation=None,
            expr_target=None,
        ),
    ],
    decorator_list=[],
    keywords=[],
    name="ConfigClass",
    expr=None,
    identifier_name=None,
)

class_squared_hinge_config_ast = ClassDef(
    bases=[Name("object", Load())],
    body=[
        Expr(
            set_value(
                "\n{tab}{doc}{tab}\n{args}".format(
                    tab=tab,
                    doc=remove_args_from_docstring(
                        function_google_tf_squared_hinge_docstring_str
                    ),
                    args="\n".join(
                        map(
                            partial(add, tab),
                            (
                                ":cvar y_true: The ground truth values. `y_true` values are expected to be -1 or 1."
                                " If binary (0 or 1) labels are provided we will convert them to -1 or 1."
                                " shape = `[batch_size, d0, .. dN]`.",
                                ":cvar y_pred: The predicted values. shape = `[batch_size, d0, .. dN]`.",
                                ":cvar return_type: Squared hinge loss values. shape = `[batch_size, d0, .. dN-1]`.",
                            ),
                        )
                    ),
                ),
            )
        ),
        AnnAssign(
            annotation=Name("object", Load()),
            simple=1,
            target=Name("y_true", Store()),
            value=set_value(None),
            expr=None,
            expr_target=None,
            expr_annotation=None,
        ),
        AnnAssign(
            annotation=Name("object", Load()),
            simple=1,
            target=Name("y_pred", Store()),
            value=set_value(None),
            expr=None,
            expr_target=None,
            expr_annotation=None,
        ),
        AnnAssign(
            annotation=Name("str", Load()),
            simple=1,
            target=Name("return_type", Store()),
            value=set_value(
                "```K.mean(math_ops.square(math_ops.maximum(1.0 - y_true * y_pred, 0.0)), axis=-1)```"
            ),
            expr=None,
            expr_target=None,
            expr_annotation=None,
        ),
        FunctionDef(
            args=arguments(
                args=[set_arg("self")],
                defaults=[],
                kw_defaults=[],
                kwarg=None,
                kwonlyargs=[],
                vararg=None,
                arg=None,
                posonlyargs=[],
            ),
            body=[
                Assign(
                    targets=[
                        Attribute(
                            Name("self", Load()),
                            "y_pred",
                            Load(),
                        )
                    ],
                    value=Call(
                        args=[Attribute(Name("self", Load()), "y_pred", Load())],
                        func=Attribute(
                            Name("ops", Load()),
                            "convert_to_tensor_v2",
                            Load(),
                        ),
                        keywords=[],
                        expr=None,
                        expr_func=None,
                    ),
                    expr=None,
                    lineno=None,
                    **maybe_type_comment,
                ),
                Assign(
                    targets=[Attribute(Name("self", Load()), "y_true", Load())],
                    value=Call(
                        args=[
                            Attribute(
                                Name("self", Load()),
                                "y_true",
                                Load(),
                            ),
                            Attribute(
                                Attribute(
                                    Name("self", Load()),
                                    "y_pred",
                                    Load(),
                                ),
                                "dtype",
                                Load(),
                            ),
                        ],
                        func=Attribute(
                            Name("math_ops", Load()),
                            "cast",
                            Load(),
                        ),
                        keywords=[],
                        expr=None,
                        expr_func=None,
                    ),
                    expr=None,
                    lineno=None,
                    **maybe_type_comment,
                ),
                Assign(
                    targets=[
                        Attribute(
                            Name("self", Load()),
                            "y_true",
                            Load(),
                        )
                    ],
                    value=Call(
                        args=[
                            Attribute(
                                Name("self", Load()),
                                "y_true",
                                Load(),
                            )
                        ],
                        func=Name("_maybe_convert_labels", Load()),
                        keywords=[],
                        expr=None,
                        expr_func=None,
                    ),
                    expr=None,
                    lineno=None,
                    **maybe_type_comment,
                ),
                Return(
                    value=Call(
                        args=[
                            Call(
                                args=[
                                    Call(
                                        args=[
                                            BinOp(
                                                set_value(1.0),
                                                Sub(),
                                                BinOp(
                                                    Attribute(
                                                        Name("self", Load()),
                                                        "y_true",
                                                        Load(),
                                                    ),
                                                    Mult(),
                                                    Attribute(
                                                        Name("self", Load()),
                                                        "y_pred",
                                                        Load(),
                                                    ),
                                                ),
                                            ),
                                            set_value(0.0),
                                        ],
                                        func=Attribute(
                                            Name("math_ops", Load()),
                                            "maximum",
                                            Load(),
                                        ),
                                        keywords=[],
                                        expr=None,
                                        expr_func=None,
                                    )
                                ],
                                func=Attribute(
                                    Name("math_ops", Load()),
                                    "square",
                                    Load(),
                                ),
                                keywords=[],
                                expr=None,
                                expr_func=None,
                            )
                        ],
                        func=Attribute(Name("K", Load()), "mean", Load()),
                        keywords=[
                            keyword(
                                arg="axis",
                                value=UnaryOp(USub(), set_value(1)),
                                identifier=None,
                            )
                        ],
                        expr=None,
                        expr_func=None,
                    ),
                    expr=None,
                ),
            ],
            decorator_list=[],
            name="__call__",
            returns=None,
            arguments_args=None,
            identifier_name=None,
            stmt=None,
            lineno=None,
            **maybe_type_comment,
        ),
    ],
    decorator_list=[],
    keywords=[],
    expr=None,
    identifier_name=None,
    name="SquaredHingeConfig",
)

# https://github.com/tensorflow/tensorflow/blob/v2.4.1/tensorflow/python/keras/callbacks.py#L1923-L2430 [- many funcs]
tensorboard_doc_str_no_args = (
    "Enable visualizations for TensorBoard.",
    "",
    "  TensorBoard is a visualization tool provided with TensorFlow.",
    "",
    "  This callback logs events for TensorBoard, including:",
    "",
    "  * Metrics summary plots",
    "  * Training graph visualization",
    "  * Activation histograms",
    "  * Sampled profiling",
    "",
    "  If you have installed TensorFlow with pip, you should be able",
    "  to launch TensorBoard from the command line:",
    "",
    "  ```",
    "  tensorboard --logdir=path_to_your_logs",
    "  ```",
    "",
    "  You can find more information about TensorBoard",
    "  [here](https://www.tensorflow.org/get_started/summaries_and_tensorboard).",
    "",
    "",
    "  Examples:",
    "",
    "  Basic usage:",
    "",
    "  ```python",
    '  tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir="./logs")',
    "  model.fit(x_train, y_train, epochs=2, callbacks=[tensorboard_callback])",
    "  # Then run the tensorboard command to view the visualizations.",
    "  ```",
    "",
    "  Custom batch-level summaries in a subclassed Model:",
    "",
    "  ```python",
    "  class MyModel(tf.keras.Model):",
    "",
    "    def build(self, _):",
    "      self.dense = tf.keras.layers.Dense(10)",
    "",
    "    def call(self, x):",
    "      outputs = self.dense(x)",
    "      tf.summary.histogram('outputs', outputs)",
    "      return outputs",
    "",
    "  model = MyModel()",
    "  model.compile('sgd', 'mse')",
    "",
    "  # Make sure to set `update_freq=N` to log a batch-level summary every N batches.",
    "  # In addition to any `tf.summary` contained in `Model.call`, metrics added in",
    "  # `Model.compile` will be logged every N batches.",
    "  tb_callback = tf.keras.callbacks.TensorBoard('./logs', update_freq=1)",
    "  model.fit(x_train, y_train, callbacks=[tb_callback])",
    "  ```",
    "",
    "  Custom batch-level summaries in a Functional API Model:",
    "",
    "  ```python",
    "  def my_summary(x):",
    "    tf.summary.histogram('x', x)",
    "    return x",
    "",
    "  inputs = tf.keras.Input(10)",
    "  x = tf.keras.layers.Dense(10)(inputs)",
    "  outputs = tf.keras.layers.Lambda(my_summary)(x)",
    "  model = tf.keras.Model(inputs, outputs)",
    "  model.compile('sgd', 'mse')",
    "",
    "  # Make sure to set `update_freq=N` to log a batch-level summary every N batches.",
    "  # In addition to any `tf.summary` contained in `Model.call`, metrics added in",
    "  # `Model.compile` will be logged every N batches.",
    "  tb_callback = tf.keras.callbacks.TensorBoard('./logs', update_freq=1)",
    "  model.fit(x_train, y_train, callbacks=[tb_callback])",
    "  ```",
    "",
    "  Profiling:",
    "",
    "  ```python",
    "  # Profile a single batch, e.g. the 5th batch.",
    "  tensorboard_callback = tf.keras.callbacks.TensorBoard(",
    "      log_dir='./logs', profile_batch=5)",
    "  model.fit(x_train, y_train, epochs=2, callbacks=[tensorboard_callback])",
    "",
    "  # Profile a range of batches, e.g. from 10 to 20.",
    "  tensorboard_callback = tf.keras.callbacks.TensorBoard(",
    "      log_dir='./logs', profile_batch=(10,20))",
    "  model.fit(x_train, y_train, epochs=2, callbacks=[tensorboard_callback])",
    "  ```",
)

tensorboard_doc_str_no_args_str = "\n".join(
    map(
        partial(strip_starting, str_to_strip="  "),
        tensorboard_doc_str_no_args,
    )
)

tensorboard_doc_str_no_args_examples_idx = tensorboard_doc_str_no_args.index(
    "  Examples:"
)

tensorboard_doc_str_args = (
    "  Args:",
    "      log_dir: the path of the directory where to save the log files to be",
    "        parsed by TensorBoard.",
    "      histogram_freq: frequency (in epochs) at which to compute activation and",
    "        weight histograms for the layers of the model. If set to 0, histograms",
    "        won't be computed. Validation data (or split) must be specified for",
    "        histogram visualizations.",
    "      write_graph: whether to visualize the graph in TensorBoard. The log file",
    "        can become quite large when write_graph is set to True.",
    "      write_images: whether to write model weights to visualize as image in",
    "        TensorBoard.",
    "      update_freq: `'batch'` or `'epoch'` or integer. When using `'batch'`,",
    "        writes the losses and metrics to TensorBoard after each batch. The same",
    "        applies for `'epoch'`. If using an integer, let's say `1000`, the",
    "        callback will write the metrics and losses to TensorBoard every 1000",
    "        batches. Note that writing too frequently to TensorBoard can slow down",
    "        your training.",
    "      profile_batch: Profile the batch(es) to sample compute characteristics.",
    "        profile_batch must be a non-negative integer or a tuple of integers.",
    "        A pair of positive integers signify a range of batches to profile.",
    "        By default, it will profile the second batch. Set profile_batch=0",
    "        to disable profiling.",
    "      embeddings_freq: frequency (in epochs) at which embedding layers will be",
    "        visualized. If set to 0, embeddings won't be visualized.",
    "      embeddings_metadata: a dictionary which maps layer name to a file name in",
    "        which metadata for this embedding layer is saved. See the",
    "        [details](",
    "          https://www.tensorflow.org/how_tos/embedding_viz/#metadata_optional)",
    "        about metadata files format. In case if the same metadata file is",
    "        used for all embedding layers, string can be passed.",
)

tensorboard_doc_str = "\n".join(
    map(
        partial(strip_starting, str_to_strip="  "),
        chain.from_iterable(
            (
                tensorboard_doc_str_no_args[:tensorboard_doc_str_no_args_examples_idx],
                "",
                tensorboard_doc_str_args,
                ("\n",),
                tensorboard_doc_str_no_args[tensorboard_doc_str_no_args_examples_idx:],
            )
        ),
    )
)
del tensorboard_doc_str_no_args_examples_idx

class_google_tf_tensorboard_str = '''
class TensorBoard(Callback, version_utils.TensorBoardVersionSelector):
  # pylint: disable=line-too-long
  r"""{tensorboard_doc_str}"""

  # pylint: enable=line-too-long

  def __init__(self,
               log_dir='logs',
               histogram_freq=0,
               write_graph=True,
               write_images=False,
               update_freq='epoch',
               profile_batch=2,
               embeddings_freq=0,
               embeddings_metadata=None,
               **kwargs):
    super(TensorBoard, self).__init__()

  def set_model(self, model):
    """Sets Keras model and writes graph if specified."""
'''.format(
    tensorboard_doc_str=tensorboard_doc_str
)

class_google_tf_tensorboard_ast = ClassDef(
    name="TensorBoard",
    bases=[
        Name("Callback", Load()),
        Attribute(Name("version_utils", Load()), "TensorBoardVersionSelector", Load()),
    ],
    keywords=[],
    body=[
        Expr(set_value(tensorboard_doc_str)),
        FunctionDef(
            name="__init__",
            args=arguments(
                posonlyargs=[],
                arg=None,
                args=list(
                    map(
                        set_arg,
                        (
                            "self",
                            "log_dir",
                            "histogram_freq",
                            "write_graph",
                            "write_images",
                            "update_freq",
                            "profile_batch",
                            "embeddings_freq",
                            "embeddings_metadata",
                        ),
                    )
                ),
                kwonlyargs=[],
                kw_defaults=[],
                kwarg=set_arg("kwargs"),
                defaults=list(
                    map(set_value, ("logs", 0, True, False, "epoch", 2, 0, None))
                ),
                vararg=None,
            ),
            body=[
                Expr(
                    Call(
                        func=Attribute(
                            Call(
                                func=Name("super", Load()),
                                args=[
                                    Name("TensorBoard", Load()),
                                    Name("self", Load()),
                                ],
                                keywords=[],
                                expr=None,
                                expr_func=None,
                            ),
                            "__init__",
                            Load(),
                        ),
                        args=[],
                        keywords=[],
                        expr=None,
                        expr_func=None,
                    )
                )
            ],
            decorator_list=[],
            arguments_args=None,
            identifier_name=None,
            stmt=None,
            lineno=None,
            returns=None,
            **maybe_type_comment,
        ),
        FunctionDef(
            name="set_model",
            args=arguments(
                posonlyargs=[],
                arg=None,
                args=list(map(set_arg, ("self", "model"))),
                kwarg=None,
                kwonlyargs=[],
                kw_defaults=[],
                defaults=[],
                vararg=None,
            ),
            body=[Expr(set_value("Sets Keras model and writes graph if specified."))],
            decorator_list=[],
            arguments_args=None,
            identifier_name=None,
            stmt=None,
            lineno=None,
            returns=None,
            **maybe_type_comment,
        ),
    ],
    decorator_list=[],
    expr=None,
    identifier_name=None,
)

class_torch_nn_l1loss_docstring = (
    "Creates a criterion that measures the mean absolute error (MAE) between each element in",
    "    the input :math:`x` and target :math:`y`.",
    "",
    "    The unreduced (i.e. with :attr:`reduction` set to ``'none'``) loss can be described as:",
    "",
    "    .. math::",
    "        \\ell(x, y) = L = \\{l_1,\\dots,l_N\\}^\\top, \\quad",
    "        l_n = \\left| x_n - y_n \\right|,",
    "",
    "    where :math:`N` is the batch size. If :attr:`reduction` is not ``'none'``",
    "    (default ``'mean'``), then:",
    "",
    "    .. math::",
    "        \\ell(x, y) =",
    "        \\begin{cases}",
    "            \\operatorname{mean}(L), & \\text{if reduction} = \\text{'mean';}\\\\",
    "            \\operatorname{sum}(L),  & \\text{if reduction} = \\text{'sum'.}",
    "        \\end{cases}",
    "",
    "    :math:`x` and :math:`y` are tensors of arbitrary shapes with a total",
    "    of :math:`n` elements each.",
    "",
    "    The sum operation still operates over all the elements, and divides by :math:`n`.",
    "",
    "    The division by :math:`n` can be avoided if one sets ``reduction = 'sum'``.",
    "",
    "    Args:",
    "        size_average (bool, optional): Deprecated (see :attr:`reduction`). By default,",
    "            the losses are averaged over each loss element in the batch. Note that for",
    "            some losses, there are multiple elements per sample. If the field :attr:`size_average`",
    "            is set to ``False``, the losses are instead summed for each minibatch. Ignored",
    "            when reduce is ``False``. Default: ``True``",
    "        reduce (bool, optional): Deprecated (see :attr:`reduction`). By default, the",
    "            losses are averaged or summed over observations for each minibatch depending",
    "            on :attr:`size_average`. When :attr:`reduce` is ``False``, returns a loss per",
    "            batch element instead and ignores :attr:`size_average`. Default: ``True``",
    "        reduction (string, optional): Specifies the reduction to apply to the output:",
    "            ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no reduction will be applied,",
    "            ``'mean'``: the sum of the output will be divided by the number of",
    "            elements in the output, ``'sum'``: the output will be summed. Note: :attr:`size_average`",
    "            and :attr:`reduce` are in the process of being deprecated, and in the meantime,",
    "            specifying either of those two args will override :attr:`reduction`. Default: ``'mean'``",
    "",
    "    Shape:",
    "        - Input: :math:`(N, *)` where :math:`*` means, any number of additional",
    "          dimensions",
    "        - Target: :math:`(N, *)`, same shape as the input",
    "        - Output: scalar. If :attr:`reduction` is ``'none'``, then",
    "          :math:`(N, *)`, same shape as the input",
    "",
    "    Examples::",
    "",
    "        >>> loss = nn.L1Loss()",
    "        >>> input = torch.randn(3, 5, requires_grad=True)",
    "        >>> target = torch.randn(3, 5)",
    "        >>> output = loss(input, target)",
    "        >>> output.backward()",
    "    ",
)
class_torch_nn_l1loss_docstring_str = "\n".join(class_torch_nn_l1loss_docstring)

class_torch_nn_l1loss_str = '''
class L1Loss(_Loss):
    r"""{class_torch_nn_l1loss_docstring_str!s}"""
    __constants__ = ['reduction']

    def __init__(self, size_average=None, reduce=None, reduction: str = 'mean') -> None:
        super(L1Loss, self).__init__(size_average, reduce, reduction)

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        return F.l1_loss(input, target, reduction=self.reduction)
'''.format(
    class_torch_nn_l1loss_docstring_str=class_torch_nn_l1loss_docstring_str
)

class_torch_nn_l1loss_ast = ClassDef(
    bases=[
        Name(
            "_Loss",
            Load(),
        )
    ],
    body=[
        Expr(set_value(class_torch_nn_l1loss_docstring_str)),
        Assign(
            targets=[Name("__constants__", Store())],
            value=List([set_value("reduction")], Load()),
            expr=None,
            lineno=None,
            **maybe_type_comment,
        ),
        FunctionDef(
            args=arguments(
                args=[
                    set_arg("self"),
                    set_arg("size_average"),
                    set_arg("reduce"),
                    set_arg(annotation=Name("str", Load()), arg="reduction"),
                ],
                defaults=list(map(set_value, (None, None, "mean"))),
                kw_defaults=[],
                kwarg=None,
                kwonlyargs=[],
                posonlyargs=[],
                vararg=None,
                arg=None,
            ),
            body=[
                Expr(
                    Call(
                        args=[
                            Name(
                                "size_average",
                                Load(),
                            ),
                            Name("reduce", Load()),
                            Name(
                                "reduction",
                                Load(),
                            ),
                        ],
                        func=Attribute(
                            Call(
                                args=[
                                    Name("L1Loss", Load()),
                                    Name(
                                        "self",
                                        Load(),
                                    ),
                                ],
                                func=Name("super", Load()),
                                keywords=[],
                                expr=None,
                                expr_func=None,
                            ),
                            "__init__",
                            Load(),
                        ),
                        keywords=[],
                        expr=None,
                        expr_func=None,
                    )
                )
            ],
            decorator_list=[],
            name="__init__",
            returns=set_value(None),
            arguments_args=None,
            identifier_name=None,
            stmt=None,
            lineno=None,
            **maybe_type_comment,
        ),
        FunctionDef(
            args=arguments(
                args=[
                    set_arg("self"),
                    set_arg(annotation=Name("Tensor", Load()), arg="input"),
                    set_arg(annotation=Name("Tensor", Load()), arg="target"),
                ],
                defaults=[],
                kw_defaults=[],
                kwarg=None,
                kwonlyargs=[],
                posonlyargs=[],
                vararg=None,
                arg=None,
            ),
            body=[
                Return(
                    value=Call(
                        args=[Name("input", Load()), Name("target", Load())],
                        func=Attribute(
                            Name("F", Load()),
                            "l1_loss",
                            Load(),
                        ),
                        keywords=[
                            keyword(
                                arg="reduction",
                                value=Attribute(
                                    Name("self", Load()),
                                    "reduction",
                                    Load(),
                                ),
                                identifier=None,
                            )
                        ],
                        expr=None,
                        expr_func=None,
                    ),
                    expr=None,
                )
            ],
            decorator_list=[],
            name="forward",
            returns=Name(
                "Tensor",
                Load(),
            ),
            arguments_args=None,
            identifier_name=None,
            stmt=None,
            lineno=None,
            **maybe_type_comment,
        ),
    ],
    decorator_list=[],
    keywords=[],
    name="L1Loss",
    expr=None,
    identifier_name=None,
)

class_torch_nn_one_cycle_lr_docstring = (
    "Sets the learning rate of each parameter group according to the",
    "    1cycle learning rate policy. The 1cycle policy anneals the learning",
    "    rate from an initial learning rate to some maximum learning rate and then",
    "    from that maximum learning rate to some minimum learning rate much lower",
    "    than the initial learning rate.",
    "    This policy was initially described in the paper `Super-Convergence:",
    "    Very Fast Training of Neural Networks Using Large Learning Rates`_.",
    "",
    "    The 1cycle learning rate policy changes the learning rate after every batch.",
    "    `step` should be called after a batch has been used for training.",
    "",
    "    This scheduler is not chainable.",
    "",
    "    Note also that the total number of steps in the cycle can be determined in one",
    "    of two ways (listed in order of precedence):",
    "",
    "    #. A value for total_steps is explicitly provided.",
    "    #. A number of epochs (epochs) and a number of steps per epoch",
    "       (steps_per_epoch) are provided.",
    "       In this case, the number of total steps is inferred by",
    "       total_steps = epochs * steps_per_epoch",
    "",
    "    You must either provide a value for total_steps or provide a value for both",
    "    epochs and steps_per_epoch.",
    "",
    "    Args:",
    "        optimizer (Optimizer): Wrapped optimizer.",
    "        max_lr (float or list): Upper learning rate boundaries in the cycle",
    "            for each parameter group.",
    "        total_steps (int): The total number of steps in the cycle. Note that",
    "            if a value is not provided here, then it must be inferred by providing",
    "            a value for epochs and steps_per_epoch.",
    "            Default: None",
    "        epochs (int): The number of epochs to train for. This is used along",
    "            with steps_per_epoch in order to infer the total number of steps in the cycle",
    "            if a value for total_steps is not provided.",
    "            Default: None",
    "        steps_per_epoch (int): The number of steps per epoch to train for. This is",
    "            used along with epochs in order to infer the total number of steps in the",
    "            cycle if a value for total_steps is not provided.",
    "            Default: None",
    "        pct_start (float): The percentage of the cycle (in number of steps) spent",
    "            increasing the learning rate.",
    "            Default: 0.3",
    "        anneal_strategy (str): {'cos', 'linear'}",
    '            Specifies the annealing strategy: "cos" for cosine annealing, "linear" for',
    "            linear annealing.",
    "            Default: 'cos'",
    "        cycle_momentum (bool): If ``True``, momentum is cycled inversely",
    "            to learning rate between 'base_momentum' and 'max_momentum'.",
    "            Default: True",
    "        base_momentum (float or list): Lower momentum boundaries in the cycle",
    "            for each parameter group. Note that momentum is cycled inversely",
    "            to learning rate; at the peak of a cycle, momentum is",
    "            'base_momentum' and learning rate is 'max_lr'.",
    "            Default: 0.85",
    "        max_momentum (float or list): Upper momentum boundaries in the cycle",
    "            for each parameter group. Functionally,",
    "            it defines the cycle amplitude (max_momentum - base_momentum).",
    "            Note that momentum is cycled inversely",
    "            to learning rate; at the start of a cycle, momentum is 'max_momentum'",
    "            and learning rate is 'base_lr'",
    "            Default: 0.95",
    "        div_factor (float): Determines the initial learning rate via",
    "            initial_lr = max_lr/div_factor",
    "            Default: 25",
    "        final_div_factor (float): Determines the minimum learning rate via",
    "            min_lr = initial_lr/final_div_factor",
    "            Default: 1e4",
    "        last_epoch (int): The index of the last batch. This parameter is used when",
    "            resuming a training job. Since `step()` should be invoked after each",
    "            batch instead of after each epoch, this number represents the total",
    "            number of *batches* computed, not the total number of epochs computed.",
    "            When last_epoch=-1, the schedule is started from the beginning.",
    "            Default: -1",
    "        verbose (bool): If ``True``, prints a message to stdout for",
    "            each update. Default: ``False``.",
    "",
    "    Example:",
    "        >>> data_loader = torch.utils.data.DataLoader(...)",
    "        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)",
    "        >>> scheduler = torch.optim.lr_scheduler.OneCycleLR(optimizer, max_lr=0.01,"
    " steps_per_epoch=len(data_loader), epochs=10)",
    "        >>> for epoch in range(10):",
    "        >>>     for batch in data_loader:",
    "        >>>         train_batch(...)",
    "        >>>         scheduler.step()",
    "",
    "",
    "    .. _Super-Convergence\\: Very Fast Training of Neural Networks Using Large Learning Rates:",
    "        https://arxiv.org/abs/1708.07120",
    "    ",
)
class_torch_nn_one_cycle_lr_docstring_str = "\n".join(
    class_torch_nn_one_cycle_lr_docstring
)[: -len(tab)]

class_torch_nn_one_cycle_lr = (
    "class OneCycleLR(_LRScheduler):",
    '    r"""{}"""'.format(class_torch_nn_one_cycle_lr_docstring_str),
    "    def __init__(self,",
    "                 optimizer,",
    "                 max_lr,",
    "                 total_steps=None,",
    "                 epochs=None,",
    "                 steps_per_epoch=None,",
    "                 pct_start=0.3,",
    "                 anneal_strategy='cos',",
    "                 cycle_momentum=True,",
    "                 base_momentum=0.85,",
    "                 max_momentum=0.95,",
    "                 div_factor=25.,",
    "                 final_div_factor=1e4,",
    "                 last_epoch=-1,",
    "                 verbose=False):",
    "",
    "        pass",
)
class_torch_nn_one_cycle_lr_str = "\n".join(class_torch_nn_one_cycle_lr)


class_torch_nn_one_cycle_lr_ast = ClassDef(
    bases=[Name("_LRScheduler", Load())],
    body=[
        Expr(set_value(class_torch_nn_one_cycle_lr_docstring_str)),
        FunctionDef(
            args=arguments(
                args=list(
                    map(
                        set_arg,
                        (
                            "self",
                            "optimizer",
                            "max_lr",
                            "total_steps",
                            "epochs",
                            "steps_per_epoch",
                            "pct_start",
                            "anneal_strategy",
                            "cycle_momentum",
                            "base_momentum",
                            "max_momentum",
                            "div_factor",
                            "final_div_factor",
                            "last_epoch",
                            "verbose",
                        ),
                    )
                ),
                defaults=list(
                    chain.from_iterable(
                        (
                            map(
                                set_value,
                                (
                                    None,
                                    None,
                                    None,
                                    0.3,
                                    "cos",
                                    True,
                                    0.85,
                                    0.95,
                                    25.0,
                                    10000.0,
                                ),
                            ),
                            iter((UnaryOp(USub(), set_value(1)), set_value(False))),
                        )
                    )
                ),
                kw_defaults=[],
                kwarg=None,
                kwonlyargs=[],
                posonlyargs=[],
                vararg=None,
                arg=None,
            ),
            body=[Pass()],
            decorator_list=[],
            name="__init__",
            returns=None,
            arguments_args=None,
            identifier_name=None,
            stmt=None,
            lineno=None,
            **maybe_type_comment,
        ),
    ],
    decorator_list=[],
    keywords=[],
    expr=None,
    identifier_name=None,
    name="OneCycleLR",
)

# From `tf.keras.losses.Reduction` @ tf-nightly:2.7.0.dev20210908, minus methods and decorator
class_reduction_v2 = ClassDef(
    name="ReductionV2",
    bases=[],
    keywords=[],
    body=[
        Expr(value=set_value(docstring_reduction_v2_str)),
        Assign(targets=[Name(id="AUTO", ctx=Store())], value=set_value("auto")),
        Assign(targets=[Name(id="NONE", ctx=Store())], value=set_value("none")),
        Assign(targets=[Name(id="SUM", ctx=Store())], value=set_value("sum")),
        Assign(
            targets=[Name(id="SUM_OVER_BATCH_SIZE", ctx=Store())],
            value=set_value("sum_over_batch_size"),
        ),
    ],
    decorator_list=[],
)

__all__ = [
    "class_ast",
    "class_ast_with_none",
    "class_google_tf_tensorboard_ast",
    "class_google_tf_tensorboard_str",
    "class_nargs_ast",
    "class_nargs_str",
    "class_reduction_v2",
    "class_squared_hinge_config_ast",
    "class_str",
    "class_torch_nn_l1loss_ast",
    "class_torch_nn_l1loss_docstring_str",
    "class_torch_nn_l1loss_str",
    "class_torch_nn_one_cycle_lr_ast",
    "class_torch_nn_one_cycle_lr_str",
    "tensorboard_doc_str_no_args_str",
    "tensorboard_doc_str",
]



"""
OpenAPI emitter function(s)
"""

from collections import deque

from cdd.compound.openapi.utils.emit_openapi_utils import (
    components_paths_from_name_model_route_id_crud,
)
from cdd.tests.mocks.json_schema import server_error_schema


def openapi(name_model_route_id_cruds):
    """
    Emit OpenAPI dict

    :param name_model_route_id_cruds: Collection of (name, model, route, id, crud)
    :type name_model_route_id_cruds: ```Iterable[NameModelRouteIdCrud]```

    :return: OpenAPI dict
    :rtype: ```dict```
    """
    paths, components = {}, {
        "requestBodies": {},
        "schemas": {
            "ServerError": {
                k: v for k, v in server_error_schema.items() if not k.startswith("$")
            }
        },
    }

    deque(
        map(
            lambda name_model_route_id_crud: components_paths_from_name_model_route_id_crud(
                components, paths, *name_model_route_id_crud
            ),
            name_model_route_id_cruds,
        ),
        maxlen=0,
    )

    return {
        "openapi": "3.0.0",
        "info": {"version": "0.0.1", "title": "REST API"},
        # "servers": [{"url": "https://example.io/v1"}],
        "components": components,
        "paths": paths,
    }


__all__ = ["openapi"]



"""
Docstring emitter.

Emits into these formats from the cdd_python common IR format:
 - [ReST docstring format (Sphinx)](
  https://sphinx-rtd-tutorial.readthedocs.io/en/latest/docstrings.html#the-sphinx-docstring-format)
 - [numpydoc docstring format](https://numpydoc.readthedocs.io/en/latest/format.html)
 - [Google's docstring format](https://google.github.io/styleguide/pyguide.html)
"""

from collections import OrderedDict
from functools import partial

from cdd.shared.docstring_utils import (
    ARG_TOKENS,
    RETURN_TOKENS,
    emit_param_str,
    header_args_footer_to_str,
    parse_docstring_into_header_args_footer,
)
from cdd.shared.pure_utils import num_of_nls, tab


def docstring(
    intermediate_repr,
    docstring_format="rest",
    purpose="function",
    word_wrap=True,
    indent_level=0,
    emit_separating_tab=True,
    emit_types=True,
    emit_original_whitespace=False,
    emit_default_doc=True,
):
    """
    Converts an IR to a docstring

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param purpose: Emit `:param` if purpose == 'function' elif purpose == 'class' then `:cvar`
    :type purpose: ```Literal['class', 'function']```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param indent_level: indentation level whence: 0=no_tabs, 1=one tab; 2=two tabs
    :type indent_level: ```int```

    :param emit_separating_tab: Whether to put a tab between :param and return and desc
    :type emit_separating_tab: ```bool```

    :param emit_types: Whether to show `:type` lines
    :type emit_types: ```bool```

    :param emit_original_whitespace: Whether to emit original whitespace or strip it out
    :type emit_original_whitespace: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: docstring
    :rtype: ```str```
    """
    # _sep = tab * indent_level
    params = "\n{maybe_nl}".format(
        maybe_nl="\n" if docstring_format == "rest" and purpose != "class" else ""
    ).join(
        (
            lambda param_lines: [getattr(ARG_TOKENS, docstring_format)[0]] + param_lines
            if param_lines and docstring_format != "rest"
            else param_lines
        )(
            list(
                map(
                    partial(
                        emit_param_str,
                        style=docstring_format,
                        purpose=purpose,
                        emit_type=emit_types,
                        emit_default_doc=emit_default_doc,
                        word_wrap=word_wrap,
                    ),
                    (intermediate_repr["params"] or OrderedDict()).items(),
                ),
            )
        )
    )

    returns = (
        (
            lambda line_: "".join(
                "{maybe_nl0_and_token}{maybe_nl1}{returns_doc}".format(
                    maybe_nl0_and_token=""
                    if docstring_format == "rest"
                    else "\n{return_token}".format(
                        return_token=getattr(RETURN_TOKENS, docstring_format)[0]
                    ),
                    maybe_nl1="" if not params or params[-1] == "\n" else "\n",
                    returns_doc=line_,
                )
            )
            if line_
            else ""
        )(
            next(
                map(
                    partial(
                        emit_param_str,
                        style=docstring_format,
                        purpose=purpose,
                        emit_type=emit_types,
                        emit_default_doc=emit_default_doc,
                        word_wrap=word_wrap,
                    ),
                    intermediate_repr["returns"].items(),
                ),
                None,
            )
        )
        if "return_type" in (intermediate_repr.get("returns") or iter(()))
        else ""
    )

    params_end_nls = num_of_nls(params, end=True)
    returns_end_nls = num_of_nls(returns, end=True)

    candidate_args_returns = "{params}{maybe_nl0}{returns}{maybe_nl1}".format(
        params=params,
        maybe_nl0="\n" if params_end_nls < 2 and returns else "",
        returns=returns,
        maybe_nl1="\n"
        if not returns and params_end_nls > 0 or returns and returns_end_nls == 0
        else "",
    )

    original_doc_str = intermediate_repr.get("_internal", {}).get(
        "original_doc_str", ""
    )
    if original_doc_str:
        header, _, footer = parse_docstring_into_header_args_footer(
            candidate_args_returns, original_doc_str
        )
        header = (
            intermediate_repr.get("doc", "") if not header and not footer else header
        )
    else:
        header, footer = intermediate_repr.get("doc", ""), ""

    candidate_doc_str = header_args_footer_to_str(
        header=header,
        args_returns="" if candidate_args_returns.isspace() else candidate_args_returns,
        footer=footer,
    )

    if not candidate_doc_str or candidate_doc_str.isspace():
        return ""

    prev_nl, next_nl = 0, candidate_doc_str.find("\n")
    current_indent, line = 0, None

    # One line only
    if next_nl == -1:
        # current_indent = count_iter_items(takewhile(str.isspace, candidate_doc_str))
        # _sep = (indent_level - current_indent) * tab
        return (
            candidate_doc_str
            if candidate_doc_str[0] == "\n"
            else "\n{_sep}{candidate_doc_str}".format(
                _sep="", candidate_doc_str=candidate_doc_str
            )
        )
    else:
        # Ignore starting newlines/whitespace only lines, keep munching until last line
        while next_nl > -1:
            line = candidate_doc_str[prev_nl:next_nl]
            if not line.isspace():
                break
            # prev_nl = next_nl
            # current_indent = count_iter_items(takewhile(str.isspace, line))

    if indent_level > current_indent:
        _tab = (indent_level - current_indent) * tab
        lines = ([line] if line else []) + candidate_doc_str[
            next_nl
            if len(candidate_doc_str) == next_nl
            or next_nl + 1 < len(candidate_doc_str)
            and candidate_doc_str[next_nl + 1] != "\n"
            else next_nl + 1 :
        ].splitlines()
        candidate_doc_str = "\n".join(
            map(
                lambda _line: "{_tab}{_line}".format(_tab=_tab, _line=_line)
                if _line or emit_separating_tab
                # and not _line.startswith(_tab)
                else _line,
                lines,
            )
        )
        if len(lines) > 1:
            candidate_doc_str = "{maybe_nl}{candidate_doc_str}{maybe_nl_tab}".format(
                maybe_nl="\n" if candidate_doc_str.startswith(_tab) else "",
                candidate_doc_str=candidate_doc_str,
                maybe_nl_tab=""
                if candidate_doc_str[-1] == "\n"
                else "\n{_tab}".format(_tab=_tab),
            )

    return candidate_doc_str


__all__ = ["docstring"]



"""
File emitter
"""

from ast import Module
from importlib import import_module
from sys import modules

from cdd.shared.source_transformer import to_code

black = (
    import_module("black")
    if "black" in modules
    else type(
        "black",
        tuple(),
        {
            "format_str": lambda src_contents, mode: src_contents,
            "Mode": lambda target_versions, line_length, is_pyi, string_normalization: None,
        },
    )
)


def file(node, filename, mode="a", skip_black=False):
    """
    Convert AST to a file

    :param node: AST node
    :type node: ```Union[Module, ClassDef, FunctionDef]```

    :param filename: emit to this file
    :type filename: ```str```

    :param mode: Mode to open the file in, defaults to append
    :type mode: ```str```

    :param skip_black: Whether to skip formatting with black
    :type skip_black: ```bool```

    :return: None
    :rtype: ```NoneType```
    """
    if not isinstance(node, Module):
        node = Module(body=[node], type_ignores=[], stmt=None)
    src = to_code(node)
    if not skip_black:
        src = black.format_str(
            src,
            mode=black.Mode(
                target_versions=set(),
                line_length=119,
                is_pyi=False,
                string_normalization=False,
            ),
        )
    with open(filename, mode) as f:
        f.write(src)


__all__ = ["file"]



"""
FastAPI route mocks
"""

from ast import (
    AsyncFunctionDef,
    Attribute,
    Call,
    Compare,
    Dict,
    If,
    Is,
    Load,
    Name,
    Return,
    arguments,
    keyword,
)

from cdd.shared.ast_utils import maybe_type_comment, set_arg, set_value

fastapi_post_create_config_str = """
@app.post(
    "/api/config",
    response_model=Config,
    responses={
        201: {
            "model": Config,
            "description": "A `Config` object.",
        },
        404: {
            "model": ServerError,
            "description": "A `ServerError` object.",
        },
    },
)
async def create_config(config: Config):
    if config is None:
        return JSONResponse(
            status_code=404,
            content={
                "error": "NotFound",
                "error_code": "0004",
                "error_description": "Config not found",
            },
        )
    return JSONResponse(status_code=201, content=config)
"""

fastapi_post_create_config_async_func = AsyncFunctionDef(
    name="create_config",
    args=arguments(
        posonlyargs=[],
        args=[set_arg(arg="config", annotation=Name(id="Config", ctx=Load()))],
        kwonlyargs=[],
        kw_defaults=[],
        defaults=[],
        vararg=None,
        kwarg=None,
        arg=None,
    ),
    body=[
        If(
            test=Compare(
                left=Name(id="config", ctx=Load()),
                ops=[Is()],
                comparators=[set_value(None)],
            ),
            body=[
                Return(
                    value=Call(
                        func=Name(id="JSONResponse", ctx=Load()),
                        args=[],
                        keywords=[
                            keyword(arg="status_code", value=set_value(404)),
                            keyword(
                                arg="content",
                                value=Dict(
                                    keys=list(
                                        map(
                                            set_value,
                                            (
                                                "error",
                                                "error_code",
                                                "error_description",
                                            ),
                                        )
                                    ),
                                    values=list(
                                        map(
                                            set_value,
                                            ("NotFound", "0004", "Config not found"),
                                        )
                                    ),
                                ),
                            ),
                        ],
                    )
                )
            ],
            orelse=[],
        ),
        Return(
            value=Call(
                func=Name(id="JSONResponse", ctx=Load()),
                args=[],
                keywords=[
                    keyword(arg="status_code", value=set_value(201)),
                    keyword(arg="content", value=Name(id="config", ctx=Load())),
                ],
            )
        ),
    ],
    decorator_list=[
        Call(
            func=Attribute(value=Name(id="app", ctx=Load()), attr="post", ctx=Load()),
            args=[set_value("/api/config")],
            keywords=[
                keyword(arg="response_model", value=Name(id="Config", ctx=Load())),
                keyword(
                    arg="responses",
                    value=Dict(
                        keys=list(map(set_value, (201, 404))),
                        values=[
                            Dict(
                                keys=[set_value("model"), set_value("description")],
                                values=[
                                    Name(id="Config", ctx=Load()),
                                    set_value("A `Config` object."),
                                ],
                            ),
                            Dict(
                                keys=[set_value("model"), set_value("description")],
                                values=[
                                    Name(id="ServerError", ctx=Load()),
                                    set_value("A `ServerError` object."),
                                ],
                            ),
                        ],
                    ),
                ),
            ],
        )
    ],
    lineno=None,
    returns=None,
    **maybe_type_comment
)


__all__ = ["fastapi_post_create_config_str", "fastapi_post_create_config_async_func"]



"""
SQLalchemy parsers

TODO
====

  - Implement update (see https://github.com/sqlalchemy/sqlalchemy/discussions/5940)
  - Implement batch CRUD

"""

import ast
from ast import AnnAssign, Assign, Call, ClassDef
from collections import OrderedDict
from inspect import getsource

import cdd.shared.parse.utils.parser_utils
from cdd.compound.openapi.utils.emit_utils import sqlalchemy_class_to_table
from cdd.docstring.parse import docstring
from cdd.shared.ast_utils import get_value
from cdd.shared.defaults_utils import extract_default
from cdd.shared.pure_utils import assert_equal
from cdd.sqlalchemy.utils.parse_utils import column_call_to_param


def sqlalchemy_table(call_or_name, parse_original_whitespace=False):
    """
    Parse out a `sqlalchemy.Table`, or a `name = sqlalchemy.Table`, into the IR

    :param call_or_name: The call to `sqlalchemy.Table` or an assignment followed by the call
    :type call_or_name: ```Union[AnnAssign, Assign, Call]```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :return: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :rtype: ```dict```
    """
    if isinstance(call_or_name, Assign):
        name, call_or_name = call_or_name.targets[0].id, call_or_name.value
    elif isinstance(call_or_name, AnnAssign):
        name, call_or_name = call_or_name.target.id, call_or_name.value
    else:
        if not isinstance(call_or_name, Call):
            call_or_name = get_value(call_or_name)
        name = get_value(call_or_name.args[0])

    # Binding should be same name as table… I guess?
    assert_equal(get_value(call_or_name.args[0]), name)

    comment = next(
        map(
            get_value,
            map(
                get_value, filter(lambda kw: kw.arg == "comment", call_or_name.keywords)
            ),
        ),
        None,
    )
    doc = next(
        map(
            get_value,
            map(get_value, filter(lambda kw: kw.arg == "doc", call_or_name.keywords)),
        ),
        None,
    )
    intermediate_repr = (
        {"type": None, "doc": "", "params": OrderedDict()}
        if comment is None and doc is None
        else docstring(
            doc or comment, parse_original_whitespace=parse_original_whitespace
        )
    )
    intermediate_repr["name"] = name
    assert isinstance(call_or_name, Call), "Expected `all` got `{node_name!r}`".format(
        node_name=type(call_or_name).__name__
    )
    assert_equal(call_or_name.func.id.rpartition(".")[2], "Table")
    assert len(call_or_name.args) > 2

    merge_ir = {
        "params": OrderedDict(map(column_call_to_param, call_or_name.args[2:])),
        "returns": None,
    }
    cdd.shared.parse.utils.parser_utils.ir_merge(
        target=intermediate_repr, other=merge_ir
    )
    if intermediate_repr["returns"] and intermediate_repr["returns"].get(
        "return_type", {}
    ).get("doc"):
        intermediate_repr["returns"]["return_type"]["doc"] = extract_default(
            intermediate_repr["returns"]["return_type"]["doc"], emit_default_doc=False
        )[0]

    return intermediate_repr


def sqlalchemy(class_def, parse_original_whitespace=False):
    """
    Parse out a `class C(Base): __tablename__=  'tbl'; dataset_name = Column(String, doc="p", primary_key=True)`,
        as constructed on an SQLalchemy declarative `Base`.

    :param class_def: A class inheriting from declarative `Base`, where `Base = sqlalchemy.orm.declarative_base()`
    :type class_def: ```Union[ClassDef]```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :return: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :rtype: ```dict```
    """

    if not isinstance(class_def, ClassDef):
        class_def = ast.parse(getsource(class_def)).body[0]
    assert isinstance(class_def, ClassDef), "Expected `ClassDef` got `{!r}`".format(
        type(class_def).__name__
    )

    return sqlalchemy_table(
        sqlalchemy_class_to_table(class_def, parse_original_whitespace)
    )


__all__ = ["sqlalchemy_table", "sqlalchemy"]



"""
Tests OpenAPI bulk
"""

from functools import partial
from os import path
from os.path import extsep
from tempfile import TemporaryDirectory
from unittest import TestCase

from cdd.compound.openapi.gen_openapi import openapi_bulk
from cdd.tests.mocks.openapi import openapi_dict_with_sql_types
from cdd.tests.mocks.routes import (
    create_route,
    destroy_route,
    read_route,
    route_mock_prelude,
)
from cdd.tests.mocks.sqlalchemy import (
    config_tbl_with_comments_str,
    sqlalchemy_imports_str,
)
from cdd.tests.utils_for_tests import unittest_main


class TestOpenApiBulk(TestCase):
    """Tests whether `openapi` can construct a `dict`"""

    def test_openapi_bulk(self) -> None:
        """
        Tests whether `openapi_bulk` produces `openapi_dict` given `model_paths` and `routes_paths`
        """
        with TemporaryDirectory() as tempdir:
            temp_dir_join = partial(path.join, tempdir)
            open(temp_dir_join("__init__{extsep}py".format(extsep=extsep)), "a").close()

            models_filename = temp_dir_join("models{extsep}py".format(extsep=extsep))
            routes_filename = temp_dir_join("routes{extsep}py".format(extsep=extsep))

            with open(models_filename, "wt") as f:
                f.write(
                    "\n".join((sqlalchemy_imports_str, config_tbl_with_comments_str))
                )

            with open(routes_filename, "wt") as f:
                f.write(
                    "\n".join(
                        (route_mock_prelude, create_route, read_route, destroy_route)
                    )
                )

            gen, gold = (
                openapi_bulk(
                    app_name="rest_api",
                    model_paths=(models_filename,),
                    routes_paths=(routes_filename,),
                ),
                openapi_dict_with_sql_types,
            )

            self.assertEqual(
                *map(
                    lambda d: d["components"]["schemas"]["Config"]["description"],
                    (gen, gold),
                )
            )

            self.assertDictEqual(gen, gold)


unittest_main()



"""
Test for the `cdd.emit` module
"""

from os import path
from unittest import TestCase

from cdd.shared.emit import EMITTERS
from cdd.shared.pure_utils import all_dunder_for_module
from cdd.tests.utils_for_tests import unittest_main


class TestEmitters(TestCase):
    """
    Tests the `cdd.emit` module magic `__all__`
    """

    def test_emitters_root(self) -> None:
        """Confirm that emitter names are up-to-date"""
        self.assertListEqual(
            EMITTERS,
            all_dunder_for_module(
                path.dirname(path.dirname(path.dirname(__file__))),
                ("sqlalchemy_table",),
            ),
        )


unittest_main()



"""
Functionality to generate classes, functions, and/or argparse functions from the input mapping
"""

import ast
from ast import Import, ImportFrom, Module
from collections import deque
from functools import partial
from inspect import getfile
from os import listdir, path

import cdd.compound.openapi.utils.emit_utils
import cdd.json_schema.emit
import cdd.shared.parse.utils.parser_utils
from cdd.compound.gen_utils import (
    file_to_input_mapping,
    gen_file,
    get_emit_kwarg,
    get_input_mapping_from_path,
    get_parser,
)
from cdd.shared.ast_utils import get_at_root
from cdd.shared.emit.utils.emitter_utils import get_emitter
from cdd.shared.pure_utils import get_module, sanitise_emit_name
from cdd.shared.source_transformer import to_code


def gen(
    name_tpl,
    input_mapping,
    parse_name,
    emit_name,
    output_filename,
    prepend=None,
    imports_from_file=None,
    emit_call=False,
    emit_default_doc=True,
    emit_and_infer_imports=False,
    decorator_list=None,
    phase=0,
    no_word_wrap=None,
):
    """
    Generate classes, functions, and/or argparse functions from the input mapping

    :param name_tpl: Template for the name, e.g., `{name}Config`.
    :type name_tpl: ```str```

    :param input_mapping: Fully-qualified module, filepath, or directory.
    :type input_mapping: ```str```

    :param parse_name: Which type to parse.
    :type parse_name: ```Literal["argparse", "class", "function", "json_schema",
                                 "pydantic", "sqlalchemy", "sqlalchemy_table", "infer"]```

    :param emit_name: Which type to generate.
    :type emit_name: ```Literal["argparse", "class", "function", "json_schema",
                                "pydantic", "sqlalchemy", "sqlalchemy_table"]```

    :param output_filename: Output file to write to
    :type output_filename: ```str```

    :param prepend: Prepend file with this. Use '\n' for newlines.
    :type prepend: ```Optional[str]```

    :param imports_from_file: Extract imports from file and append to `output_file`.
        If module or other symbol path given, resolve file then use it.
    :type imports_from_file: ```Optional[str]```

    :param emit_call: Whether to emit a `__call__` method from the `_internal` IR subdict
    :type emit_call: ```bool```

    :param emit_and_infer_imports: Whether to emit and infer imports at the top of the generated code
    :type emit_and_infer_imports: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param phase: Which phase to run through. E.g., SQLalchemy may require multiple phases to resolve foreign keys
    :type phase: ```int```

    :param no_word_wrap: Whether word-wrap is disabled (on emission).
    :type no_word_wrap: ```Optional[Literal[True]]```
    """
    extra_symbols = {}
    if phase > 0 and emit_name in frozenset(("sqlalchemy", "sqlalchemy_table")):
        if phase == 1:
            return (
                cdd.compound.openapi.utils.emit_utils.update_with_imports_from_columns(
                    output_filename
                )
            )
        elif phase == 2:
            return cdd.compound.openapi.utils.emit_utils.update_fk_for_file(
                output_filename
            )
        else:
            raise NotImplementedError("phase {}".format(phase))
    elif imports_from_file is None:
        imports = ""
    else:
        if prepend:
            prepend_imports = get_at_root(
                ast.parse(prepend.strip()), (Import, ImportFrom)
            )

            # def rewrite_typings(node):
            #     """
            #     Python < 3.8 must use `typings_extensions` for `Literal`
            #
            #     :param node: import node
            #     :type node: ```Union[Import, ImportFrom]```
            #
            #     :return: The import potentially rewritten or None
            #     :rtype: ```Optional[Union[Import, ImportFrom]]```
            #     """
            #     if isinstance(node, ImportFrom) and node.module == "typing":
            #         len_names = len(node.names)
            #         if len_names == 1 and node.names[0].name == "Literal":
            #             rewrite_typings.found_literal = True
            #             return None
            #         else:
            #             node.names = list(
            #                 filter(
            #                     None,
            #                     map(
            #                         lambda _alias: None
            #                         if _alias.name == "Literal"
            #                         else _alias,
            #                         node.names,
            #                     ),
            #                 )
            #             )
            #             if len(node.names) != len_names:
            #                 rewrite_typings.found_literal = True
            #     return node
            #
            # rewrite_typings.found_literal = False
            # prepend_imports = list(filter(None, map(rewrite_typings, prepend_imports)))
            # if rewrite_typings.found_literal:
            #     prepend_imports.append(
            #         ImportFrom(
            #             level=0,
            #             module="typing_extensions"
            #             if sys.version_info[:2] < (3, 8)
            #             else "typing",
            #             names=[alias(asname=None, name="Literal")],
            #             lineno=None,
            #             col_offset=None,
            #         )
            #     )

            eval(
                compile(
                    to_code(
                        ast.fix_missing_locations(
                            Module(body=prepend_imports, stmt=None, type_ignores=[])
                        )
                    ),
                    filename="<string>",
                    mode="exec",
                ),
                extra_symbols,
            )
            # This leaks to the global scope
            globals().update(extra_symbols)
        with open(
            imports_from_file
            if path.isfile(imports_from_file)
            else getfile(get_module(imports_from_file, extra_symbols=extra_symbols)),
            "rt",
        ) as f:
            imports = "".join(
                map(to_code, get_at_root(ast.parse(f.read()), (Import, ImportFrom)))
            )

    module_path, _, symbol_name = input_mapping.rpartition(".")

    emit_name = sanitise_emit_name(emit_name)
    if path.isfile(input_mapping):
        input_mapping = file_to_input_mapping(input_mapping, parse_name)
    elif path.isdir(input_mapping):
        _input_mapping = {}
        deque(
            map(
                _input_mapping.update,
                map(
                    partial(
                        file_to_input_mapping,
                        parse_name=parse_name,
                    ),
                    map(partial(path.join, input_mapping), listdir(input_mapping)),
                ),
            ),
            maxlen=0,
        )
        input_mapping = _input_mapping
        del _input_mapping
    else:
        input_mod = get_module(module_path, extra_symbols=extra_symbols)
        input_mapping = (
            getattr(input_mod, symbol_name)
            if hasattr(input_mod, symbol_name)
            else get_input_mapping_from_path(emit_name, module_path, symbol_name)
        )
    input_mapping_it = (
        input_mapping.items() if hasattr(input_mapping, "items") else input_mapping
    )

    return (
        cdd.json_schema.emit.json_schema_file(
            {
                name: get_emitter(emit_name)(
                    get_parser(node, parse_name)(node),
                    emit_default_doc=emit_default_doc,
                    word_wrap=no_word_wrap is None,
                    **get_emit_kwarg(
                        decorator_list, emit_call, emit_name, name_tpl, name
                    ),
                )
                for name, node in input_mapping_it
            },
            output_filename,
        )
        if emit_name == "json_schema"
        else gen_file(
            name_tpl,
            input_mapping_it,
            parse_name,
            emit_name,
            output_filename,
            prepend,
            emit_call,
            emit_and_infer_imports,
            emit_default_doc,
            decorator_list,
            no_word_wrap,
            imports,
        )
    )


__all__ = ["gen"]



"""
Utility functions for `cdd.parse.function`
"""

import ast
from ast import Return, Tuple
from collections import OrderedDict

from cdd.shared.ast_utils import get_value
from cdd.shared.pure_utils import rpartial
from cdd.shared.source_transformer import to_code


def _interpolate_return(function_def, intermediate_repr):
    """
    Interpolate the return value into the IR.

    :param function_def: function definition
    :type function_def: ```FunctionDef```

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :return: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :rtype: ```dict```
    """
    return_ast = next(
        filter(rpartial(isinstance, Return), function_def.body[::-1]), None
    )
    if return_ast is not None and return_ast.value is not None:
        if intermediate_repr.get("returns") is None:
            intermediate_repr["returns"] = OrderedDict((("return_type", {}),))

        if (
            "typ" in intermediate_repr["returns"]["return_type"]
            and "[" not in intermediate_repr["returns"]["return_type"]["typ"]
        ):
            del intermediate_repr["returns"]["return_type"]["typ"]
        intermediate_repr["returns"]["return_type"]["default"] = (
            lambda default: "({})".format(default)
            if isinstance(return_ast.value, Tuple)
            and (not default.startswith("(") or not default.endswith(")"))
            else (
                lambda default_: default_
                if isinstance(
                    default_, (str, int, float, complex, ast.Num, ast.Str, ast.Constant)
                )
                else "```{}```".format(default)
            )(get_value(get_value(return_ast)))
        )(to_code(return_ast.value).rstrip("\n"))
    if hasattr(function_def, "returns") and function_def.returns is not None:
        intermediate_repr["returns"] = intermediate_repr.get("returns") or OrderedDict(
            (("return_type", {}),)
        )
        intermediate_repr["returns"]["return_type"]["typ"] = to_code(
            function_def.returns
        ).rstrip("\n")

    return intermediate_repr


__all__ = ["_interpolate_return"]



# -*- coding: utf-8 -*-

"""
setup.py implementation, interesting because it parsed the first __init__.py and
    extracts the `__author__` and `__version__`
"""

from ast import Assign, Constant, Str, parse
from distutils.sysconfig import get_python_lib
from functools import partial
from operator import attrgetter
from os import path
from os.path import extsep

from setuptools import find_packages, setup

package_name = "cdd"

with open(
    path.join(path.dirname(__file__), "README{extsep}md".format(extsep=extsep)),
    "rt",
    encoding="utf-8",
) as fh:
    long_description = fh.read()


def to_funcs(*paths):
    """
    Produce function tuples that produce the local and install dir, respectively.

    :param paths: one or more str, referring to relative folder names
    :type paths: ```*paths```

    :return: 2 functions
    :rtype: ```Tuple[Callable[Optional[List[str]], str], Callable[Optional[List[str]], str]]```
    """
    return (
        partial(path.join, path.dirname(__file__), package_name, *paths),
        partial(path.join, get_python_lib(prefix=""), package_name, *paths),
    )


def main():
    """Main function for setup.py; this actually does the installation"""
    with open(
        path.join(
            path.abspath(path.dirname(__file__)),
            package_name,
            "__init__{extsep}py".format(extsep=extsep),
        )
    ) as f:
        parsed_init = parse(f.read())

    __author__, __version__, __description__ = map(
        lambda node: node.value if isinstance(node, Constant) else node.s,
        filter(
            lambda node: isinstance(node, (Constant, Str)),
            map(
                attrgetter("value"),
                filter(lambda node: isinstance(node, Assign), parsed_init.body),
            ),
        ),
    )

    setup(
        name="python-" + package_name,
        author=__author__,
        author_email="807580+SamuelMarks@users.noreply.github.com",
        version=__version__,
        description=__description__,
        long_description=long_description,
        long_description_content_type="text/markdown",
        install_requires=["pyyaml"],
        test_suite=package_name + ".tests",
        packages=find_packages(),
        package_dir={package_name: package_name},
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "License :: OSI Approved",
            "License :: OSI Approved :: Apache Software License",
            "License :: OSI Approved :: MIT License",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: Implementation",
            "Programming Language :: Python :: Implementation :: CPython",
            "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
            "Topic :: Software Development",
            "Topic :: Software Development :: Build Tools",
            "Topic :: Software Development :: Code Generators",
            "Topic :: Software Development :: Compilers",
            "Topic :: Software Development :: Pre-processors",
            "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        ],
        python_requires=">=3.6",
        url="https://github.com/offscale/cdd-python",
    )


def setup_py_main():
    """Calls main if `__name__ == '__main__'`"""
    if __name__ == "__main__":
        main()


setup_py_main()



"""
Tests for the utils that is used by the SQLalchemy parsers
"""

from copy import deepcopy
from unittest import TestCase

from cdd.sqlalchemy.utils.parse_utils import (
    column_call_name_manipulator,
    column_call_to_param,
    get_pk_and_type,
    get_table_name,
)
from cdd.tests.mocks.ir import intermediate_repr_node_pk
from cdd.tests.mocks.json_schema import config_schema
from cdd.tests.mocks.sqlalchemy import (
    config_decl_base_ast,
    dataset_primary_key_column_assign,
    node_fk_call,
)
from cdd.tests.utils_for_tests import unittest_main


class TestParseSqlAlchemyUtils(TestCase):
    """
    Tests the SQLalchemy parser utilities
    """

    def test_column_call_to_param_pk(self) -> None:
        """
        Tests that `parse.sqlalchemy.utils.column_call_to_param` works with PK
        """

        gold_name, gold_param = (
            lambda _name: (
                _name,
                {
                    "default": config_schema["properties"][_name]["default"],
                    "typ": "str",
                    "doc": config_schema["properties"][_name]["description"],
                    "x_typ": {"sql": {"type": "String"}},
                },
            )
        )("dataset_name")
        gen_name, gen_param = column_call_to_param(
            column_call_name_manipulator(
                deepcopy(dataset_primary_key_column_assign.value), "add", gold_name
            )
        )
        self.assertEqual(gold_name, gen_name)
        self.assertDictEqual(gold_param, gen_param)

    def test_column_call_to_param_fk(self) -> None:
        """
        Tests that `parse.sqlalchemy.utils.column_call_to_param` works with FK
        """
        gen_name, gen_param = column_call_to_param(deepcopy(node_fk_call))
        gold_name, gold_param = (
            lambda _name: (_name, deepcopy(intermediate_repr_node_pk["params"][_name]))
        )("primary_element")
        self.assertEqual(gold_name, gen_name)
        self.assertDictEqual(gold_param, gen_param)

    def test_get_pk_and_type(self) -> None:
        """
        Tests get_pk_and_type
        """
        self.assertEqual(
            get_pk_and_type(config_decl_base_ast), ("dataset_name", "String")
        )
        no_pk = deepcopy(config_decl_base_ast)
        del no_pk.body[2]
        self.assertIsNone(get_pk_and_type(no_pk))

    def test_get_table_name(self) -> None:
        """
        Tests `get_table_name`
        """
        self.assertEqual(get_table_name(config_decl_base_ast), "config_tbl")
        no_table_name = deepcopy(config_decl_base_ast)
        del no_table_name.body[1]
        self.assertEqual(get_table_name(no_table_name), "Config")


unittest_main()



"""
Tests for the Intermediate Representation produced by the JSON schema parser
"""

from unittest import TestCase

import cdd.json_schema.emit
import cdd.json_schema.parse
from cdd.tests.mocks.ir import intermediate_repr_no_default_sql_doc
from cdd.tests.mocks.json_schema import config_schema
from cdd.tests.utils_for_tests import unittest_main


class TestParseJsonSchema(TestCase):
    """
    Tests whether the intermediate representation is consistent when parsed from different inputs.

    IR is a dictionary of form:
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    """

    def test_from_json_schema(self) -> None:
        """
        Tests that `parse.json_schema` produces `intermediate_repr_no_default_sql_doc` properly
        """
        self.assertDictEqual(
            cdd.json_schema.parse.json_schema(config_schema),
            intermediate_repr_no_default_sql_doc,
        )


unittest_main()



"""
Utility functions for `cdd.emit.docstring`
"""

from cdd.shared.ast_utils import NoneStr
from cdd.shared.defaults_utils import extract_default
from cdd.shared.pure_utils import simple_types, unquote


def interpolate_defaults(
    param, default_search_announce=None, require_default=False, emit_default_doc=True
):
    """
    Correctly set the 'default' and 'doc' parameters

    :param param: Name, dict with keys: 'typ', 'doc', 'default'
    :type param: ```Tuple[str, dict]```

    :param default_search_announce: Default text(s) to look for. If None, uses default specified in default_utils.
    :type default_search_announce: ```Optional[Union[str, Iterable[str]]]```

    :param require_default: Whether a default is required, if not found in doc, infer the proper default from type
    :type require_default: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Name, dict with keys: 'typ', 'doc', 'default'
    :rtype: ```Tuple[str, dict]```
    """
    name, _param = param
    del param
    if "doc" in _param:
        doc, default = extract_default(
            _param["doc"],
            typ=_param.get("typ"),
            default_search_announce=default_search_announce,
            emit_default_doc=emit_default_doc,
        )
        _param["doc"] = doc
        if default is not None:
            _param["default"] = unquote(default)
    if require_default and _param.get("default") is None:
        # if (
        #     "typ" in _param
        #     and _param["typ"] not in frozenset(("Any", "object"))
        #     and not _param["typ"].startswith("Optional")
        # ):
        #     _param["typ"] = "Optional[{}]".format(_param["typ"])
        _param["default"] = (
            simple_types[_param["typ"]]
            if _param.get("typ", memoryview) in simple_types
            else NoneStr
        )

    return name, _param


__all__ = ["interpolate_defaults"]



"""
Tests for `cdd.emit.docstring`
"""

import ast
from ast import Expr, FunctionDef, arguments
from copy import deepcopy
from functools import partial
from unittest import TestCase

import cdd.argparse_function.emit
import cdd.argparse_function.parse
import cdd.class_.emit
import cdd.class_.parse
import cdd.docstring.emit
import cdd.function.emit
import cdd.function.parse
import cdd.json_schema.emit
import cdd.shared.emit.file
import cdd.sqlalchemy.emit
from cdd.shared.pure_utils import omit_whitespace, reindent
from cdd.tests.mocks.classes import class_ast
from cdd.tests.mocks.docstrings import (
    docstring_google_str,
    docstring_google_tf_ops_losses__safe_mean_str,
    docstring_no_default_no_nl_str,
    docstring_no_nl_str,
    docstring_numpydoc_str,
)
from cdd.tests.mocks.ir import (
    function_google_tf_ops_losses__safe_mean_ir,
    intermediate_repr,
)
from cdd.tests.mocks.methods import function_google_tf_ops_losses__safe_mean_ast
from cdd.tests.utils_for_tests import unittest_main


class TestEmitDocstring(TestCase):
    """Tests emission"""

    def test_to_docstring(self) -> None:
        """
        Tests whether `docstring` produces indented `docstring_str` given `class_ast`
        """
        self.assertEqual(
            cdd.docstring.emit.docstring(
                cdd.class_.parse.class_(class_ast), emit_default_doc=True
            ),
            reindent(docstring_no_nl_str, 1),
        )

    def test_to_docstring_emit_default_doc_false(self) -> None:
        """
        Tests whether `docstring` produces `docstring_str` given `class_ast`
        """
        ir = cdd.class_.parse.class_(class_ast)
        self.assertEqual(
            cdd.docstring.emit.docstring(ir, emit_default_doc=False),
            reindent(docstring_no_default_no_nl_str, 1),
        )

    def test_to_numpy_docstring(self) -> None:
        """
        Tests whether `docstring` produces `docstring_numpydoc_str` when `docstring_format` is 'numpydoc'
        """
        self.assertEqual(
            docstring_numpydoc_str,
            cdd.docstring.emit.docstring(
                deepcopy(intermediate_repr), docstring_format="numpydoc"
            ),
        )

    def test_to_google_docstring(self) -> None:
        """
        Tests whether `docstring` produces `docstring_google_str` when `docstring_format` is 'google'
        """
        self.assertEqual(
            docstring_google_str,
            cdd.docstring.emit.docstring(
                deepcopy(intermediate_repr), docstring_format="google"
            ),
        )

    def test_to_google_docstring_no_types(self) -> None:
        """
        Tests whether a Google docstring is correctly generated sans types
        """

        self.assertEqual(
            *map(
                omit_whitespace,
                (
                    docstring_google_tf_ops_losses__safe_mean_str,
                    cdd.docstring.emit.docstring(
                        deepcopy(function_google_tf_ops_losses__safe_mean_ir),
                        docstring_format="google",
                        emit_original_whitespace=True,
                        emit_default_doc=False,
                        word_wrap=True,
                    ),
                ),
            )
        )

    def test_to_docstring_use_original_when_whitespace_only_changes(self) -> None:
        """
        Tests whether original docstring is used when whitespace only changes are made
        """

        self.assertEqual(
            *map(
                partial(ast.get_docstring, clean=True),
                map(
                    lambda doc_str: FunctionDef(
                        name="_",
                        args=arguments(
                            posonlyargs=[],
                            args=[],
                            kwonlyargs=[],
                            kw_defaults=[],
                            defaults=[],
                            vararg=None,
                            kwarg=None,
                        ),
                        body=[Expr(doc_str)],
                        decorator_list=[],
                        lineno=None,
                        returns=None,
                    ),
                    (
                        cdd.docstring.emit.docstring(
                            cdd.function.parse.function(
                                function_google_tf_ops_losses__safe_mean_ast
                            ),
                            docstring_format="google",
                            emit_original_whitespace=True,
                            emit_default_doc=False,
                        ),
                        docstring_google_tf_ops_losses__safe_mean_str,
                    ),
                ),
            )
        )


unittest_main()



"""
Emit constant strings with interpolated values for route generation
"""

from cdd.routes.emit.bottle_constants_utils import (
    create_helper_variants,
    create_route_variants,
    delete_route_variants,
    read_route_variants,
)


def create(app, name, route, variant=2):
    """
    Create the `create` route

    :param app: Variable name (Bottle App)
    :type app: ```str```

    :param name: Name of entity
    :type name: ```str```

    :param route: The path of the resource
    :type route: ```str```

    :param variant: Number of variant
    :type variant: ```int```

    :return: Create route variant with interpolated values
    :rtype: ```str```
    """
    return create_route_variants[variant].format(app=app, name=name, route=route)


def create_util(name, route, variant=1):
    """
    Create utility function that the `create` emitter above uses

    :param name: Name of entity
    :type name: ```str```

    :param route: The path of the resource
    :type route: ```str```

    :param variant: Number of variant
    :type variant: ```int```

    :return: Create route variant with interpolated values
    :rtype: ```str```
    """
    return create_helper_variants[variant].format(name=name, route=route)


def read(app, name, route, primary_key, variant=0):
    """
    Create the `read` route

    :param app: Variable name (Bottle App)
    :type app: ```str```

    :param name: Name of entity
    :type name: ```str```

    :param route: The path of the resource
    :type route: ```str```

    :param primary_key: The id
    :type primary_key: ```Any```

    :param variant: Number of variant
    :type variant: ```int```

    :return: Create route variant with interpolated values
    :rtype: ```str```
    """
    return read_route_variants[variant].format(
        app=app, name=name, route=route, id=primary_key
    )


def destroy(app, name, route, primary_key, variant=0):
    """
    Create the `destroy` route

    :param app: Variable name (Bottle App)
    :type app: ```str```

    :param name: Name of entity
    :type name: ```str```

    :param route: The path of the resource
    :type route: ```str```

    :param primary_key: The id
    :type primary_key: ```Any```

    :param variant: Number of variant
    :type variant: ```int```

    :return: Create route variant with interpolated values
    :rtype: ```str```
    """
    return delete_route_variants[variant].format(
        app=app, name=name, route=route, id=primary_key
    )


__all__ = ["create", "create_util", "read", "destroy"]



"""
Argparse function parser
"""

from ast import Assign, FunctionDef, Return, Tuple, get_docstring
from collections import OrderedDict
from functools import partial
from itertools import filterfalse
from operator import setitem

from cdd.argparse_function.utils.emit_utils import _parse_return, parse_out_param
from cdd.shared.ast_utils import (
    get_function_type,
    get_value,
    is_argparse_add_argument,
    is_argparse_description,
)
from cdd.shared.docstring_parsers import parse_docstring


def argparse_ast(
    function_def,
    function_type=None,
    function_name=None,
    parse_original_whitespace=False,
    word_wrap=False,
):
    """
    Converts an argparse AST to our IR

    :param function_def: AST of argparse function_def
    :type function_def: ```FunctionDef```

    :param function_type: Type of function, static is static or global method, others just become first arg
    :type function_type: ```Literal['self', 'cls', 'static']```

    :param function_name: name of function_def
    :type function_name: ```str```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :return: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :rtype: ```dict```
    """
    assert isinstance(
        function_def, FunctionDef
    ), "Expected `FunctionDef` got `{node_name!r}`".format(
        node_name=type(function_def).__name__
    )

    doc_string = get_docstring(function_def, clean=parse_original_whitespace)
    intermediate_repr = {
        "name": function_name or function_def.name,
        "type": function_type or get_function_type(function_def),
        "doc": "",
        "params": OrderedDict(),
    }
    ir = parse_docstring(
        doc_string,
        word_wrap=word_wrap,
        emit_default_doc=True,
        parse_original_whitespace=parse_original_whitespace,
    )

    # Whether a default is required, if not found in doc, infer the proper default from type
    require_default = False

    # Parse all relevant nodes from function body
    body = function_def.body if doc_string is None else function_def.body[1:]
    for node in body:
        if is_argparse_add_argument(node):
            name, _param = parse_out_param(
                node,
                emit_default_doc=False,  # require_default=require_default
            )
            (
                intermediate_repr["params"][name].update
                if name in intermediate_repr["params"]
                else partial(setitem, intermediate_repr["params"], name)
            )(_param)
            if not require_default and _param.get("default") is not None:
                require_default = True
        elif isinstance(node, Assign) and is_argparse_description(node):
            intermediate_repr["doc"] = get_value(node.value)
        elif isinstance(node, Return) and isinstance(node.value, Tuple):
            intermediate_repr["returns"] = OrderedDict(
                (
                    _parse_return(
                        node,
                        intermediate_repr=ir,
                        function_def=function_def,
                        emit_default_doc=False,
                    ),
                )
            )

    inner_body = list(
        filterfalse(
            is_argparse_description,
            filterfalse(is_argparse_add_argument, body),
        )
    )
    if inner_body:
        intermediate_repr["_internal"] = {
            "original_doc_str": doc_string
            if parse_original_whitespace
            else get_docstring(function_def, clean=False),
            "body": inner_body,
            "from_name": function_def.name,
            "from_type": "static",
        }

    return intermediate_repr


__all__ = ["argparse_ast"]



"""
Tests for the Intermediate Representation produced by the function parser
"""

from collections import OrderedDict
from copy import deepcopy
from unittest import TestCase

import cdd.argparse_function.emit
import cdd.class_.parse
import cdd.function.parse
import cdd.json_schema.parse
from cdd.shared.ast_utils import get_value
from cdd.shared.pure_utils import PY_GTE_3_8, paren_wrap_code
from cdd.tests.mocks.ir import (
    function_adder_ir,
    function_google_tf_ops_losses__safe_mean_ir,
    method_complex_args_variety_ir,
)
from cdd.tests.mocks.methods import (
    function_adder_ast,
    function_adder_str,
    function_default_complex_default_arg_ast,
    function_google_tf_ops_losses__safe_mean_ast,
    method_complex_args_variety_ast,
    method_complex_args_variety_str,
)
from cdd.tests.utils_for_tests import inspectable_compile, unittest_main


class TestParseFunction(TestCase):
    """
    Tests whether the intermediate representation is consistent when parsed from different inputs.

    IR is a dictionary of form:
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    """

    def test_from_function(self) -> None:
        """
        Tests that cdd.parse.function.function produces properly
        """
        gen_ir = cdd.function.parse.function(function_default_complex_default_arg_ast)
        gold_ir = {
            "name": "call_peril",
            "params": OrderedDict(
                (
                    (
                        "dataset_name",
                        {"default": "mnist", "typ": "str"},
                    ),
                    (
                        "writer",
                        {
                            "default": "```{}```".format(
                                paren_wrap_code(
                                    get_value(
                                        function_default_complex_default_arg_ast.args.defaults[
                                            1
                                        ]
                                    )
                                )
                            ),
                        },
                    ),
                )
            ),
            "returns": None,
            "type": "static",
        }

        del gen_ir["_internal"]  # Not needed for this test
        self.assertDictEqual(
            gen_ir,
            gold_ir,
        )

    def test_from_function_kw_only(self) -> None:
        """
        Tests that cdd.parse.function.function produces properly from function with only keyword arguments
        """
        gen_ir = cdd.function.parse.function(function_adder_ast, function_type="static")
        del gen_ir["_internal"]  # Not needed for this test
        self.assertDictEqual(
            gen_ir,
            function_adder_ir,
        )

    def test_from_function_in_memory(self) -> None:
        """
        Tests that cdd.parse.function.function produces properly from a function in memory of current interpreter
        """

        def foo(a=5, b=6):
            """
            the foo function

            :param a: the a value
            :param b: the b value

            """

        self.assertIsNone(foo(5, 6))

        ir = cdd.function.parse.function(foo)
        del ir["_internal"]  # Not needed for this test
        self.assertDictEqual(
            ir,
            {
                "doc": "the foo function",
                "name": "TestParseFunction.test_from_function_in_memory.<locals>.foo",
                "params": OrderedDict(
                    (
                        ("a", {"default": 5, "doc": "the a value", "typ": "int"}),
                        ("b", {"default": 6, "doc": "the b value", "typ": "int"}),
                    )
                ),
                "returns": None,
                "type": "static",
            },
        )

    def test_from_method_in_memory(self) -> None:
        """
        Tests that `parse.function` produces properly from a function in memory of current interpreter with:
        - kw only args;
        - default args;
        - annotations
        - required;
        - unannotated;
        - splat
        """

        method_complex_args_variety_with_imports_str = (
            "from sys import stdout\n"
            "from {package} import Literal\n"
            "{body}".format(
                package="typing" if PY_GTE_3_8 else "typing_extensions",
                body=method_complex_args_variety_str,
            )
        )
        call_cliff = getattr(
            inspectable_compile(method_complex_args_variety_with_imports_str),
            "call_cliff",
        )

        ir = cdd.function.parse.function(call_cliff)
        del ir["_internal"]  # Not needed for this test

        # This is a hack because JetBrains wraps stdout
        self.assertIn(
            type(ir["params"]["writer"]["default"]).__name__,
            frozenset(("EncodedFile", "FlushingStringIO", "TextIOWrapper")),
        )

        # This extra typ is copied, for now. TODO: Update AST-level parser to set types when defaults are given.

        del ir["params"]["writer"]["typ"]
        ir["params"]["writer"]["default"] = "```{}```".format(paren_wrap_code("stdout"))

        self.assertDictEqual(
            ir,
            method_complex_args_variety_ir,
        )

    def test_from_method_in_memory_return_complex(self) -> None:
        """
        Tests that `parse.function` produces properly from a function in memory of current interpreter with:
        - complex return type
        - kwonly args
        """

        method_complex_args_variety_with_imports_str = (
            "from sys import stdout\n"
            "from {package} import Literal\n"
            "{body}".format(
                package="typing" if PY_GTE_3_8 else "typing_extensions",
                body=function_adder_str,
            )
        )
        add_6_5 = getattr(
            inspectable_compile(method_complex_args_variety_with_imports_str),
            "add_6_5",
        )

        ir = cdd.function.parse.function(add_6_5)
        del ir["_internal"]  # Not needed for this test

        self.assertDictEqual(
            ir,
            function_adder_ir,
        )

    def test_from_method_complex_args_variety(self) -> None:
        """
        Tests that `parse.function` produces correctly with:
        - kw only args;
        - default args;
        - annotations
        - required;
        - unannotated;
        - splat
        """
        gen_ir = cdd.function.parse.function(method_complex_args_variety_ast)
        del gen_ir["_internal"]  # Not needed for this test
        self.assertDictEqual(
            gen_ir,
            method_complex_args_variety_ir,
        )

    def test_from_function_google_tf_ops_losses__safe_mean_ast(self) -> None:
        """Tests IR from function_google_tf_ops_losses__safe_mean_ast"""
        ir = cdd.function.parse.function(function_google_tf_ops_losses__safe_mean_ast)
        _internal = ir.pop("_internal")
        for key in "original_doc_str", "body":
            del _internal[key]
        self.assertDictEqual(
            _internal, {"from_name": "_safe_mean", "from_type": "static"}
        )
        self.assertEqual(
            ir["returns"]["return_type"]["doc"],
            function_google_tf_ops_losses__safe_mean_ir["returns"]["return_type"][
                "doc"
            ],
        )
        self.assertDictEqual(ir, function_google_tf_ops_losses__safe_mean_ir)

        no_body = deepcopy(function_google_tf_ops_losses__safe_mean_ast)
        del no_body.body[1:]
        ir = cdd.function.parse.function(no_body)
        del ir["_internal"]
        gold = deepcopy(function_google_tf_ops_losses__safe_mean_ir)
        gold["returns"]["return_type"] = {
            "doc": function_google_tf_ops_losses__safe_mean_ir["returns"][
                "return_type"
            ]["doc"]
        }
        self.assertDictEqual(ir, gold)


unittest_main()



"""
Tests for `cdd.emit.pydantic`
"""

from unittest import TestCase

import cdd.pydantic.emit
from cdd.tests.mocks.ir import pydantic_ir
from cdd.tests.mocks.pydantic import pydantic_class_cls_def
from cdd.tests.utils_for_tests import run_ast_test, unittest_main


class TestEmitPyDantic(TestCase):
    """Tests emission"""

    def test_to_pydantic(self) -> None:
        """
        Tests whether `pydantic` produces `pydantic_class_cls_def` given `pydantic_ir`
        """
        run_ast_test(
            self,
            gen_ast=cdd.pydantic.emit.pydantic(pydantic_ir),
            gold=pydantic_class_cls_def,
        )


unittest_main()



"""
Tests for `cdd.emit.class_`
"""

from ast import FunctionDef
from unittest import TestCase

import cdd.argparse_function.emit
import cdd.argparse_function.parse
import cdd.class_.emit
import cdd.docstring.emit
import cdd.docstring.parse
import cdd.function.emit
import cdd.function.parse
import cdd.json_schema.emit
import cdd.shared.emit.file
import cdd.sqlalchemy.emit
from cdd.shared.ast_utils import annotate_ancestry, find_in_ast
from cdd.shared.pure_utils import rpartial
from cdd.tests.mocks.argparse import argparse_func_action_append_ast, argparse_func_ast
from cdd.tests.mocks.classes import class_ast, class_nargs_ast
from cdd.tests.mocks.docstrings import docstring_no_nl_str
from cdd.tests.mocks.methods import class_with_method_and_body_types_ast
from cdd.tests.utils_for_tests import reindent_docstring, run_ast_test, unittest_main


class TestEmitClass(TestCase):
    """Tests emission"""

    def test_to_class_from_argparse_ast(self) -> None:
        """
        Tests whether `class_` produces `class_ast` given `argparse_func_ast`
        """

        ir = cdd.argparse_function.parse.argparse_ast(argparse_func_ast)
        gen_ast = cdd.class_.emit.class_(
            ir,
            emit_default_doc=True,
            class_name="ConfigClass",
        )

        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=class_ast,
        )

    def test_to_class_from_argparse_action_append_ast(self) -> None:
        """
        Tests whether a class from an argparse function with `nargs` set
        """
        run_ast_test(
            self,
            cdd.class_.emit.class_(
                cdd.argparse_function.parse.argparse_ast(
                    argparse_func_action_append_ast
                ),
                class_name="ConfigClass",
            ),
            gold=class_nargs_ast,
        )

    def test_to_class_from_docstring_str(self) -> None:
        """
        Tests whether `class_` produces `class_ast` given `docstring_str`
        """
        run_ast_test(
            self,
            cdd.class_.emit.class_(
                cdd.docstring.parse.docstring(
                    docstring_no_nl_str, emit_default_doc=True
                ),
                emit_default_doc=True,
                class_name="ConfigClass",
            ),
            gold=class_ast,
        )

    def test_from_class_with_body_in_method_to_method_with_body(self) -> None:
        """Tests if this can make the roundtrip from a full function to a full function"""
        annotate_ancestry(class_with_method_and_body_types_ast)

        function_def = reindent_docstring(
            next(
                filter(
                    rpartial(isinstance, FunctionDef),
                    class_with_method_and_body_types_ast.body,
                )
            )
        )

        ir = cdd.function.parse.function(
            find_in_ast(
                "C.function_name".split("."),
                class_with_method_and_body_types_ast,
            ),
        )
        gen_ast = cdd.function.emit.function(
            ir,
            emit_default_doc=False,
            function_name="function_name",
            function_type="self",
            indent_level=0,
            emit_separating_tab=True,
            emit_as_kwonlyargs=False,
        )

        # emit.file(gen_ast, os.path.join(os.path.dirname(__file__),
        #           "delme{extsep}py".format(extsep=extsep)), mode="wt")

        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=function_def,
        )


unittest_main()



"""
Function parser
"""

import ast
from ast import AnnAssign, Assign, FunctionDef, get_docstring
from collections import OrderedDict
from copy import deepcopy
from functools import partial
from inspect import getsource
from itertools import cycle, filterfalse, islice
from types import FunctionType

import cdd.docstring.parse
import cdd.shared.parse.utils.parser_utils
from cdd.function.utils.parse_utils import _interpolate_return
from cdd.shared.ast_utils import NoneStr, func_arg2param, get_function_type
from cdd.shared.docstring_parsers import _set_name_and_type
from cdd.shared.pure_utils import rpartial


def function(
    function_def,
    infer_type=False,
    parse_original_whitespace=False,
    word_wrap=True,
    function_type=None,
    function_name=None,
):
    """
    Converts a method to our IR

    :param function_def: AST node for function definition
    :type function_def: ```Union[FunctionDef, FunctionType]```

    :param infer_type: Whether to try inferring the typ (from the default)
    :type infer_type: ```bool```

    :param parse_original_whitespace: Whether to parse original whitespace or strip it out
    :type parse_original_whitespace: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param function_type: Type of function, static is static or global method, others just become first arg
    :type function_type: ```Literal['self', 'cls', 'static']```

    :param function_name: name of function_def
    :type function_name: ```str```

    :return: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :rtype: ```dict```
    """
    if isinstance(function_def, FunctionType):
        # Dynamic function, i.e., this isn't source code; and is in your memory
        ir = cdd.shared.parse.utils.parser_utils._inspect(
            function_def,
            function_name,
            parse_original_whitespace=parse_original_whitespace,
            word_wrap=word_wrap,
        )
        parsed_source = ast.parse(getsource(function_def).lstrip()).body[0]
        original_doc_str = ast.get_docstring(
            parsed_source, clean=parse_original_whitespace
        )
        body = (
            parsed_source.body if original_doc_str is None else parsed_source.body[1:]
        )
        ir["_internal"] = {
            "original_doc_str": original_doc_str
            if parse_original_whitespace
            else ast.get_docstring(parsed_source, clean=False),
            "body": list(filterfalse(rpartial(isinstance, (AnnAssign, Assign)), body)),
            "from_name": parsed_source.name,
            "from_type": "cls",
        }
        return ir

    assert isinstance(
        function_def, FunctionDef
    ), "Expected `FunctionDef` got `{node_name!r}`".format(
        node_name=type(function_def).__name__
    )
    assert (
        function_name is None or function_def.name == function_name
    ), "Expected {function_name!r} got {function_def_name!r}".format(
        function_name=function_name, function_def_name=function_def.name
    )

    found_type = get_function_type(function_def)

    # Read docstring
    doc_str = (
        get_docstring(function_def, clean=parse_original_whitespace)
        if isinstance(function_def, FunctionDef)
        else None
    )

    function_def = deepcopy(function_def)
    function_def.args.args = (
        function_def.args.args if found_type == "static" else function_def.args.args[1:]
    )

    if doc_str is None:
        intermediate_repr = {
            "name": function_name or function_def.name,
            "params": OrderedDict(),
            "returns": None,
            "_internal": {},
        }
    else:
        intermediate_repr = cdd.docstring.parse.docstring(
            doc_str.replace(":cvar", ":param"),
            parse_original_whitespace=parse_original_whitespace,
            infer_type=infer_type,
        )
        intermediate_repr["_internal"] = {
            "original_doc_str": (
                doc_str
                if parse_original_whitespace
                else get_docstring(function_def, clean=False)
                if isinstance(function_def, FunctionDef)
                else None
            )
        }

    intermediate_repr.update(
        {
            "name": function_name or function_def.name,
            "type": function_type or found_type,
        }
    )

    intermediate_repr["_internal"].update(
        {
            "from_name": function_def.name,
            "from_type": found_type,
        }
    )
    function_def.body = function_def.body if doc_str is None else function_def.body[1:]
    if function_def.body:
        intermediate_repr["_internal"]["body"] = function_def.body

    params_to_append = OrderedDict()
    if (
        hasattr(function_def.args, "kwarg")
        and function_def.args.kwarg
        and function_def.args.kwarg.arg in intermediate_repr["params"]
    ):
        _param = intermediate_repr["params"].pop(function_def.args.kwarg.arg)
        assert "typ" in _param
        _param["default"] = NoneStr
        params_to_append[function_def.args.kwarg.arg] = _param
        del _param

    # Set defaults

    # Fill with `None`s when no default is given to make the `zip` below it work cleanly
    for args, defaults in (
        ("args", "defaults"),
        ("kwonlyargs", "kw_defaults"),
    ):
        diff = abs(
            len(getattr(function_def.args, args))
            - len(getattr(function_def.args, defaults))
        )
        if diff:
            setattr(
                function_def.args,
                defaults,
                list(islice(cycle((None,)), diff))
                + getattr(function_def.args, defaults),
            )
    cdd.shared.parse.utils.parser_utils.ir_merge(
        intermediate_repr,
        {
            "params": OrderedDict(
                (
                    func_arg2param(
                        getattr(function_def.args, args)[idx],
                        default=getattr(function_def.args, defaults)[idx],
                    )
                    for args, defaults in (
                        ("args", "defaults"),
                        ("kwonlyargs", "kw_defaults"),
                    )
                    for idx in range(len(getattr(function_def.args, args)))
                )
            ),
            "returns": None,
        },
    )

    intermediate_repr["params"].update(params_to_append)
    intermediate_repr["params"] = OrderedDict(
        map(
            partial(_set_name_and_type, infer_type=infer_type, word_wrap=word_wrap),
            intermediate_repr["params"].items(),
        )
    )

    # Convention - the final top-level `return` is the default
    intermediate_repr = _interpolate_return(function_def, intermediate_repr)
    if "return_type" in (intermediate_repr.get("returns") or iter(())):
        intermediate_repr["returns"] = OrderedDict(
            map(
                partial(_set_name_and_type, infer_type=infer_type, word_wrap=word_wrap),
                intermediate_repr["returns"].items(),
            )
        )
    return intermediate_repr


__all__ = ["function"]



"""
Functions to handle default parameterisation
"""

import ast
from ast import literal_eval
from collections import OrderedDict
from contextlib import suppress
from copy import deepcopy
from functools import partial
from itertools import takewhile
from operator import contains, eq

from cdd.shared.pure_utils import (
    PY_GTE_3_9,
    count_iter_items,
    location_within,
    none_types,
    quote,
    simple_types,
)

NoneStr = "```(None)```" if PY_GTE_3_9 else "```None```"


def ast_parse_fix(s):
    """
    Hack to resolve unbalanced parentheses SyntaxError acquired from PyTorch parsing
    TODO: remove

    :param s: String to parse
    :type s: ```str```

    :return: Value
    """
    # return ast.parse(s).body[0].value
    balanced = (s.count("[") + s.count("]")) & 1 == 0
    return ast.parse(s if balanced else "{}]".format(s)).body[0].value


def needs_quoting(typ):
    """
    Figures out whether values with this type need quoting

    :param typ: The type
    :type typ: ```Optional[str]```

    :return: Whether the type needs quoting
    :rtype: ```bool```
    """
    if typ is None or typ.startswith("*"):
        return False
    elif typ == "str":
        return True
    elif typ == "Optional[str]":
        return True

    typ = typ.replace("\n", "").strip()
    parsed_typ_ast = ast_parse_fix(typ)
    if isinstance(parsed_typ_ast, ast.Name):
        return parsed_typ_ast.id == "str"

    return any(
        filter(
            lambda node: isinstance(node, ast.Str)
            or isinstance(node, ast.Constant)
            and type(node.value).__name__ == "str"
            or isinstance(node, ast.Name)
            and node.id == "str",
            ast.walk(parsed_typ_ast),
        )
    )


def extract_default(
    line,
    rstrip_default=True,
    default_search_announce=None,
    typ=None,
    emit_default_doc=True,
):
    """
    Extract the a tuple of (doc, default) from a doc line

    :param line: Example - "dataset. Defaults to mnist"
    :type line: ```str```

    :param rstrip_default: Whether to rstrip whitespace, newlines, and '.' from the default
    :type rstrip_default: ```bool```

    :param default_search_announce: Default text(s) to look for. If None, uses default specified in default_utils.
    :type default_search_announce: ```Optional[Union[str, Iterable[str]]]```

    :param typ: The type of the default value, useful to disambiguate `25` the float from  `25` the float
    :type typ: ```Optional[str]```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Example - ("dataset. Defaults to mnist", "mnist") if emit_default_doc else ("dataset", "mnist")
    :rtype: Tuple[str, Optional[str]]
    """
    if line is None:
        return line, line

    default_search_announce_paren, default_search_announce = (
        lambda _default_search_announce: (
            map(partial(str.format, "({}"), _default_search_announce),
            _default_search_announce,
        )
    )(
        ("defaults to ", "defaults to\n", "Default value is ", "Default:")
        if default_search_announce is None
        else (
            (default_search_announce,)
            if isinstance(default_search_announce, str)
            else default_search_announce
        )
    )

    _start_idx, _end_idx, default_end_offset = None, None, None
    for idx, _default_search_announce in enumerate(
        (default_search_announce_paren, default_search_announce)
    ):
        _start_idx, _end_idx, _found = location_within(
            line,
            _default_search_announce,
            cmp=lambda a, b: eq(*map(str.casefold, (a, b))),
        )

        if idx == 0:
            if _start_idx != -1:
                _start_idx += 1  # eat '('
                default_end_offset = (
                    -1 if line[-1] == ")" else -2 if line[-2:] == ")." else 0
                )  # eat ')', ').'
                break
        elif _start_idx < 0:
            return line, None

    default = ""
    par = {"{": 0, "[": 0, "(": 0, ")": 0, "]": 0, "}": 0}
    sub_l = line[_end_idx:default_end_offset]
    sub_l_len = len(sub_l)
    for idx, ch in enumerate(sub_l):
        if (
            ch == "."
            and (idx == (sub_l_len - 1) or not (sub_l[idx + 1]).isdigit())
            and not sum(par.values())
        ):
            break
        elif ch in par:
            par[ch] += 1
        default += ch

    start_rest_offset = _end_idx + len(default)

    default = default.strip(" \t`")

    return _parse_out_default_and_doc(
        _start_idx,
        start_rest_offset,
        default,
        line,
        rstrip_default,
        typ,
        default_end_offset,
        emit_default_doc,
    )


def _parse_out_default_and_doc(
    _start_idx,
    start_rest_offset,
    default,
    line,
    rstrip_default,
    typ,
    default_end_offset,
    emit_default_doc,
):
    """
    Internal function to parse the default and extract out the doc iff `emit_default_doc is False`

    :param _start_idx: The start index to look from
    :type _start_idx: ```int```

    :param start_rest_offset: The start index to look from, for the rest that's appended
    :type start_rest_offset: ```int```

    :param default: The currently parsed out default, could be the end form, could parse into something more specific
    :type default: ```Any```

    :param line: Example - "dataset. Defaults to mnist"
    :type line: ```str```

    :param rstrip_default: Whether to rstrip whitespace, newlines, and '.' from the default
    :type rstrip_default: ```bool```

    :param typ: The type of the default value, useful to disambiguate `25` the float from  `25` the float
    :type typ: ```Optional[str]```

    :param default_end_offset: Set to -1 if one parenthesis, -2 if )., and 0 if none
    :type default_end_offset: ```int```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Example - ("dataset. Defaults to mnist", "mnist") if emit_default_doc else ("dataset", "mnist")
    :rtype: Tuple[str, Optional[str]]
    """
    if typ is not None and typ in simple_types and default not in none_types:
        lit = (
            ast.AST()
            if typ != "str"
            and any(
                map(
                    partial(contains, frozenset(("*", "^", "&", "|", "$", "@", "!"))),
                    default,
                )
            )
            else literal_eval("({default})".format(default=default))
        )
        default = (
            "```{default}```".format(default=default)
            if isinstance(lit, ast.AST)
            else {
                "bool": bool,
                "int": int,
                "float": float,
                "complex": complex,
                "str": str,
            }[typ](lit)
        )
    elif default.isdecimal():
        default = int(default)
    elif default in frozenset(("True", "False")):
        default = literal_eval(default)
    else:
        with suppress(ValueError):
            default = float(default)
    if emit_default_doc:
        return line, default
    else:
        stop_tokens = frozenset((" ", "\t", "\n", "\n", "."))
        end = line[: _start_idx - 1]
        extra_offset = int(end[-1] in frozenset((" ", "\t", "\n", "\n")) if end else 0)

        if rstrip_default:
            offset = count_iter_items(
                takewhile(
                    partial(contains, stop_tokens),
                    line[start_rest_offset:],
                )
            )
            start_rest_offset += offset

        fst = line[: _start_idx - 1 - extra_offset]
        rest = line[
            start_rest_offset : (-extra_offset if extra_offset > 0 else None)
            if default_end_offset is None
            else default_end_offset
        ]
        return (
            fst + rest,
            default,
        )


def remove_defaults_from_intermediate_repr(intermediate_repr, emit_default_prop=True):
    """
    Remove "Default of" text from IR

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param emit_default_prop: Whether to emit default property
    :type emit_default_prop: ```bool```

    :return: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :rtype: ```dict```
    """
    ir = deepcopy(intermediate_repr)

    remove_default_from_param = partial(
        _remove_default_from_param, emit_default_prop=emit_default_prop
    )
    ir.update(
        {
            key: OrderedDict(map(remove_default_from_param, ir[key].items()))
            for key in ("params", "returns")
        }
    )
    return ir


def _remove_default_from_param(param, emit_default_prop=True):
    """
    Remove default from param iff emit_default_prop is False

    :param param: Name, dict with keys: 'typ', 'doc', 'default'
    :type param: ```Tuple[str, dict]```

    :param emit_default_prop: Whether to emit default property
    :type emit_default_prop: ```bool```

    :return: Name, dict with keys: 'typ', 'doc', 'default'
    :rtype: ```Tuple[str, dict]```
    """
    name, _param = param
    del param
    doc, default = extract_default(_param["doc"], emit_default_doc=False)
    _param.update({"doc": doc, "default": default})
    if default is None or not emit_default_prop:
        del _param["default"]
    return name, _param


def set_default_doc(param, emit_default_doc=True):
    """
    Emit param with 'doc' set to include 'Defaults'

    :param param: Name, dict with keys: 'typ', 'doc', 'default'
    :type param: ```Tuple[str, dict]```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Same shape as input but with Default append to doc.
    :rtype: ```dict```
    """
    name, _param = param
    del param
    # if param is None: param = {"doc": "", "typ": "Any"}
    if _param is None or "doc" not in _param:
        return name, _param
    has_defaults = "Defaults" in _param["doc"] or "defaults" in _param["doc"]

    if has_defaults and not emit_default_doc:
        # Remove the default text
        _param["doc"] = extract_default(
            _param["doc"], emit_default_doc=emit_default_doc
        )[0]
    elif "default" in _param and not has_defaults and emit_default_doc:
        # if _param["default"] == NoneStr: _param["default"] = None
        if _param["default"] is not None or not name.endswith("kwargs"):
            _param["doc"] = "{doc} Defaults to {default}".format(
                doc=(
                    _param["doc"]
                    if _param["doc"][-1] in frozenset((".", ","))
                    else "{doc}.".format(doc=_param["doc"])
                ),
                default=(
                    quote(_param["default"])
                    if (
                        needs_quoting(_param.get("typ"))
                        and (
                            len(_param["default"]) < 2
                            or not _param["default"].startswith("`")
                            or not _param["default"].endswith("`")
                        )
                        if isinstance(_param["default"], str)
                        else True
                    )
                    else _param["default"]
                ),
            )

    return name, _param


__all__ = [
    "extract_default",
    "needs_quoting",
    "remove_defaults_from_intermediate_repr",
    "set_default_doc",
]



"""
Argparse emitter
"""

import ast
from ast import (
    Assign,
    Attribute,
    Expr,
    FunctionDef,
    Load,
    Name,
    Return,
    Store,
    Tuple,
    arguments,
)
from collections import OrderedDict
from functools import partial
from itertools import chain

from cdd.docstring.emit import docstring
from cdd.shared.ast_utils import (
    get_value,
    maybe_type_comment,
    param2argparse_param,
    set_arg,
    set_value,
)
from cdd.shared.emit.utils.emitter_utils import get_internal_body
from cdd.shared.pure_utils import code_quoted, fill, identity, none_types


def argparse_function(
    intermediate_repr,
    emit_default_doc=False,
    function_name="set_cli_args",
    function_type="static",
    wrap_description=False,
    word_wrap=True,
    docstring_format="rest",
):
    """
    Convert to an argparse FunctionDef

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param function_name: name of function_def
    :type function_name: ```str```

    :param function_type: Type of function, static is static or global method, others just become first arg
    :type function_type: ```Literal['self', 'cls', 'static']```

    :param wrap_description: Whether to word-wrap the description. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type wrap_description: ```bool```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :return:  AST node for function definition which constructs argparse
    :rtype: ```FunctionDef```
    """
    function_name = function_name or intermediate_repr["name"]
    function_type = function_type or intermediate_repr["type"]
    internal_body = get_internal_body(
        target_name=function_name,
        target_type=function_type,
        intermediate_repr=intermediate_repr,
    )

    return FunctionDef(
        args=arguments(
            args=[set_arg("argument_parser")],
            # None if function_type in frozenset((None, "static"))
            # else set_arg(function_type),
            defaults=[],
            kw_defaults=[],
            kwarg=None,
            kwonlyargs=[],
            posonlyargs=[],
            vararg=None,
            arg=None,
        ),
        body=list(
            chain.from_iterable(
                (
                    iter(
                        (
                            Expr(
                                set_value(
                                    docstring(
                                        {
                                            "doc": "Set CLI arguments",
                                            "params": OrderedDict(
                                                (
                                                    (
                                                        "argument_parser",
                                                        {
                                                            "doc": "argument parser",
                                                            "typ": "ArgumentParser",
                                                        },
                                                    ),
                                                )
                                            ),
                                            "returns": OrderedDict(
                                                (
                                                    (
                                                        "return_type",
                                                        {
                                                            "doc": "argument_parser, {returns_doc}".format(
                                                                returns_doc=intermediate_repr[
                                                                    "returns"
                                                                ][
                                                                    "return_type"
                                                                ][
                                                                    "doc"
                                                                ]
                                                            )
                                                            if intermediate_repr[
                                                                "returns"
                                                            ]["return_type"].get("doc")
                                                            else "argument_parser",
                                                            "typ": "Tuple[ArgumentParser, {typ}]".format(
                                                                typ=intermediate_repr[
                                                                    "returns"
                                                                ]["return_type"]["typ"]
                                                            ),
                                                        }
                                                        if "return_type"
                                                        in (
                                                            (
                                                                intermediate_repr or {}
                                                            ).get("returns")
                                                            or iter(())
                                                        )
                                                        and intermediate_repr[
                                                            "returns"
                                                        ]["return_type"].get("typ")
                                                        not in none_types
                                                        else {
                                                            "doc": "argument_parser",
                                                            "typ": "ArgumentParser",
                                                        },
                                                    ),
                                                ),
                                            ),
                                        },
                                        docstring_format=docstring_format,
                                        word_wrap=word_wrap,
                                        indent_level=1,
                                    )
                                )
                            ),
                            Assign(
                                targets=[
                                    Attribute(
                                        Name("argument_parser", Load()),
                                        "description",
                                        Store(),
                                    )
                                ],
                                value=set_value(
                                    (fill if wrap_description else identity)(
                                        intermediate_repr["doc"]
                                    )
                                ),
                                lineno=None,
                                expr=None,
                                **maybe_type_comment
                            ),
                        )
                    ),
                    filter(
                        None,
                        (
                            *(
                                (
                                    map(
                                        partial(
                                            param2argparse_param,
                                            word_wrap=word_wrap,
                                            emit_default_doc=emit_default_doc,
                                        ),
                                        intermediate_repr["params"].items(),
                                    )
                                )
                                if "params" in intermediate_repr
                                else ()
                            ),
                            *(
                                internal_body[
                                    2
                                    if len(internal_body) > 1
                                    and isinstance(internal_body[1], Assign)
                                    and internal_body[1].targets[0].id
                                    == "argument_parser"
                                    else 1 :
                                ]
                                if internal_body
                                and isinstance(internal_body[0], Expr)
                                and isinstance(get_value(internal_body[0].value), str)
                                else internal_body
                            ),
                            None
                            if internal_body and isinstance(internal_body[-1], Return)
                            else (
                                Return(
                                    value=Tuple(
                                        ctx=Load(),
                                        elts=[
                                            Name("argument_parser", Load()),
                                            set_value(
                                                intermediate_repr["returns"][
                                                    "return_type"
                                                ]["default"]
                                            )
                                            if code_quoted(
                                                intermediate_repr["returns"][
                                                    "return_type"
                                                ]["default"]
                                            )
                                            else ast.parse(
                                                intermediate_repr["returns"][
                                                    "return_type"
                                                ]["default"]
                                            )
                                            .body[0]
                                            .value,
                                        ],
                                        expr=None,
                                    ),
                                    expr=None,
                                )
                                if "default"
                                in (
                                    intermediate_repr.get("returns")
                                    or {"return_type": iter(())}
                                )["return_type"]
                                else Return(
                                    value=Name("argument_parser", Load()), expr=None
                                )
                            ),
                        ),
                    ),
                )
            )
        ),
        decorator_list=[],
        name=function_name,
        returns=None,
        lineno=None,
        arguments_args=None,
        identifier_name=None,
        stmt=None,
        **maybe_type_comment
    )


__all__ = ["argparse_function"]



""" Tests for parser_utils """

from collections import OrderedDict
from copy import deepcopy
from unittest import TestCase
from unittest.mock import patch

import cdd.shared.parse.utils.parser_utils
from cdd.class_.utils.parse_utils import get_source
from cdd.shared.ast_utils import set_value
from cdd.tests.mocks import imports_header
from cdd.tests.mocks.argparse import argparse_func_ast, argparse_func_str
from cdd.tests.mocks.classes import class_ast, class_str
from cdd.tests.mocks.methods import (
    function_default_complex_default_arg_ast,
    method_complex_args_variety_str,
)
from cdd.tests.mocks.sqlalchemy import (
    config_decl_base_ast,
    config_decl_base_str,
    config_tbl_with_comments_ast,
    sqlalchemy_imports_str,
)
from cdd.tests.utils_for_tests import inspectable_compile, unittest_main


class TestParserUtils(TestCase):
    """Test class for parser_utils"""

    def test_ir_merge_empty(self) -> None:
        """Tests for `ir_merge` when both are empty"""
        target = {"params": OrderedDict(), "returns": None}
        other = {"params": OrderedDict(), "returns": None}
        self.assertDictEqual(
            cdd.shared.parse.utils.parser_utils.ir_merge(target, other),
            target,
        )

    def test_ir_merge_other_empty(self) -> None:
        """Tests for `ir_merge` when only non-target is empty"""
        target = {
            "params": OrderedDict(
                (("something", {}),),
            ),
            "returns": None,
        }
        other = {"params": OrderedDict(), "returns": None}
        self.assertDictEqual(
            cdd.shared.parse.utils.parser_utils.ir_merge(target, other),
            target,
        )

    def test_ir_merge_same_len(self) -> None:
        """Tests for `ir_merge` when target and non-target have same size"""
        target = {
            "params": OrderedDict(
                (("something", {"typ": "str"}),),
            ),
            "returns": None,
        }
        other = {
            "params": OrderedDict(
                (("something", {"doc": "neat"}),),
            ),
            "returns": None,
        }
        self.assertDictEqual(
            cdd.shared.parse.utils.parser_utils.ir_merge(deepcopy(target), other),
            {
                "params": OrderedDict(
                    (("something", {"doc": "neat", "typ": "str"}),),
                ),
                "returns": None,
            },
        )

    def test_ir_merge_same_len_returns(self) -> None:
        """Tests for `ir_merge` when target and non-target have same size and a return"""
        target = {
            "params": OrderedDict(),
            "returns": OrderedDict(
                (
                    (
                        "return_type",
                        {"typ": "str"},
                    ),
                )
            ),
        }
        other = {
            "params": OrderedDict(),
            "returns": OrderedDict(
                (
                    (
                        "return_type",
                        {"doc": "so stringy"},
                    ),
                )
            ),
        }
        self.assertDictEqual(
            cdd.shared.parse.utils.parser_utils.ir_merge(deepcopy(target), other),
            {
                "params": OrderedDict(),
                "returns": OrderedDict(
                    (("return_type", {"typ": "str", "doc": "so stringy"}),)
                ),
            },
        )

    def test__join_non_none_returns_early(self) -> None:
        """Tests that `_join_non_none` returns early"""
        empty_str_dict = {"": ""}
        self.assertDictEqual(
            cdd.shared.parse.utils.parser_utils._join_non_none(
                primacy={}, other=empty_str_dict
            ),
            empty_str_dict,
        )
        self.assertDictEqual(
            cdd.shared.parse.utils.parser_utils._join_non_none(
                primacy=empty_str_dict, other={}
            ),
            empty_str_dict,
        )

    def test_infer_argparse_ast(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `argparse_ast`
        """
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(argparse_func_ast), "argparse_ast"
        )

    def test_infer_memory_argparse_ast(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `argparse_ast`
        """
        set_cli_args = getattr(
            inspectable_compile(argparse_func_str),
            "set_cli_args",
        )
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(set_cli_args), "argparse_ast"
        )

    def test_infer_docstring(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `docstring`
        """
        self.assertEqual(cdd.shared.parse.utils.parser_utils.infer(""), "docstring")
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(set_value("")), "docstring"
        )

    def test_infer_class(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `class_`
        """
        self.assertEqual(cdd.shared.parse.utils.parser_utils.infer(class_ast), "class_")

    def test_infer_memory_class(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `class_`
        """
        set_cli_args = getattr(
            inspectable_compile(imports_header + class_str),
            "ConfigClass",
        )
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(set_cli_args), "class_"
        )

    def test_infer_function(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `function`
        """
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(
                function_default_complex_default_arg_ast
            ),
            "function",
        )

    def test_infer_memory_function(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `function`
        """
        call_cliff = getattr(
            inspectable_compile(
                "\n".join(
                    (imports_header, "stdout = None", method_complex_args_variety_str)
                )
            ),
            "call_cliff",
        )
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(call_cliff), "function"
        )

    def test_infer_sqlalchemy_table(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `sqlalchemy_table`
        """
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(config_tbl_with_comments_ast),
            "sqlalchemy_table",
        )

    # def test_infer_memory_sqlalchemy_table(self) -> None:
    #     """
    #     Test `infer` can figure out the right parser name when its expected to be `sqlalchemy_table`
    #     """
    #     config_tbl = getattr(
    #         inspectable_compile("\n".join((
    #             sqlalchemy_imports_str,
    #             "metadata = None", config_tbl_str,
    #         ))),
    #         "config_tbl",
    #     )
    #     self.assertEqual(infer(config_tbl), "sqlalchemy_table")

    def test_infer_sqlalchemy(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `sqlalchemy`
        """
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(config_decl_base_ast),
            "sqlalchemy",
        )

    def test_infer_memory_sqlalchemy(self) -> None:
        """
        Test `infer` can figure out the right parser name when its expected to be `sqlalchemy`
        """
        Config = getattr(
            inspectable_compile(
                "\n".join(
                    (sqlalchemy_imports_str, "Base = object", config_decl_base_str)
                )
            ),
            "Config",
        )
        self.assertEqual(
            cdd.shared.parse.utils.parser_utils.infer(Config), "sqlalchemy"
        )

    def test_get_source_raises(self) -> None:
        """Tests that `get_source` raises an exception"""
        with self.assertRaises(TypeError):
            get_source(None)

        def raise_os_error(_):
            """raise_OSError"""
            raise OSError

        with patch("inspect.getsourcelines", raise_os_error), self.assertRaises(
            OSError
        ):
            get_source(min)

        with patch("inspect.getsourcefile", lambda _: None):
            self.assertIsNone(get_source(raise_os_error))


unittest_main()



""" Tests for CLI openapi subparser (__main__.py) """

from unittest import TestCase
from unittest.mock import patch

from cdd.tests.utils_for_tests import mock_function, run_cli_test, unittest_main


class TestOpenApi(TestCase):
    """Test class for __main__.py"""

    def test_gen_routes_fails(self) -> None:
        """Tests CLI interface failure cases"""
        run_cli_test(
            self,
            ["openapi", "--wrong"],
            exit_code=2,
            output="the following arguments are required: --model-paths, --routes-paths\n",
        )

    def test_openapi(self) -> None:
        """Tests CLI interface gets all the way to the `openapi` call without error"""

        with patch("cdd.__main__.openapi_bulk", mock_function):
            self.assertTrue(
                run_cli_test(
                    self,
                    [
                        "openapi",
                        "--app-name",
                        "app",
                        "--model-paths",
                        "cdd.tests.mocks.sqlalchemy",
                        "--routes-paths",
                        "cdd.tests.mocks.routes",
                    ],
                    exit_code=None,
                    output=None,
                ),
            )


unittest_main()



"""
Tests for AST equality
"""

import ast
from unittest import TestCase

from cdd.tests.mocks.argparse import argparse_func_ast, argparse_func_str
from cdd.tests.mocks.classes import class_ast, class_str
from cdd.tests.utils_for_tests import reindent_docstring, run_ast_test, unittest_main


class TestAstEquality(TestCase):
    """
    Tests whether the AST generated matches the mocked one expected
    """

    def test_argparse_func(self) -> None:
        """Tests whether the `argparse_func_str` correctly produces `argparse_func_ast`"""
        run_ast_test(
            self,
            *map(
                reindent_docstring,
                (ast.parse(argparse_func_str).body[0], argparse_func_ast),
            )
        )

    def test_class(self) -> None:
        """Tests whether the `class_str` correctly produces `class_ast`"""
        run_ast_test(self, class_str, class_ast)


unittest_main()



""" Tests for default utils """

from unittest import TestCase

from cdd.shared.defaults_utils import extract_default, set_default_doc
from cdd.tests.utils_for_tests import unittest_main


class TestDefaultUtils(TestCase):
    """Test class for default utils"""

    def test_extract_default(self) -> None:
        """Tests that `extract_default` produces the expected output"""
        sample = "This defaults to foo."
        self.assertTupleEqual(extract_default(sample), (sample, "foo"))
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=False), ("This", "foo")
        )

    def test_extract_default_middle(self) -> None:
        """Tests that `extract_default` produces the expected output"""
        sample = "Why would you. Have this defaults to something. In the middle?"
        default = "something"
        self.assertTupleEqual(extract_default(sample), (sample, default))

        self.assertTupleEqual(
            extract_default(sample, rstrip_default=False, emit_default_doc=False),
            ("Why would you. Have this. In the middle?", default),
        )
        self.assertTupleEqual(
            extract_default(sample, rstrip_default=True, emit_default_doc=False),
            ("Why would you. Have thisIn the middle?", default),
        )

    def test_extract_default_with_dot(self) -> None:
        """Tests that `extract_default` works when there is a `.` in the default"""
        sample = "This. defaults to (np.empty(0), np.empty(0))"
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=False),
            ("This.", "(np.empty(0), np.empty(0))"),
        )

    def test_extract_default_with_int(self) -> None:
        """Tests that `extract_default` works for an integer default"""
        sample = "learning rate. Defaults to 0001."
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=False),
            ("learning rate.", 1),
        )

    def test_extract_default_with_float(self) -> None:
        """Tests that `extract_default` works when there is a `.` in the default referring to a decimal place"""
        sample = "learning rate. Defaults to 0.001."
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=False),
            ("learning rate.", 0.001),
        )

    def test_extract_default_with_bool(self) -> None:
        """Tests that `extract_default` works for an integer default"""
        sample = (
            "Boolean. Whether to apply AMSGrad variant of this algorithm from"
            'the paper "On the Convergence of Adam and beyond". Defaults to `True`.'
        )
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=True),
            (sample, True),
        )

    def test_extract_default_with_parens(self) -> None:
        """Tests that `extract_default` works when wrapped in parentheses"""
        sample = "learning rate (default: 1)"
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=True),
            (sample, 1),
        )

        sample = (
            "tolerance_change (float): termination tolerance on function\n"
            "             value/parameter changes (default: 1e-9)."
        )
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=True),
            (sample, 1e-9),
        )

    def test_extract_default_with_many_parens(self) -> None:
        """Tests that `extract_default` works when default parses to an AST type"""
        sample = (
            "betas (Tuple[float, float], optional): coefficients used for computing\n"
            "        running averages of gradient and its square (default: (0.9, 0.999))"
        )
        default = "(0.9, 0.999)"
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=True),
            (sample, default),
        )
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=False),
            (
                "betas (Tuple[float, float], optional): coefficients used for computing\n"
                "        running averages of gradient and its square",
                default,
            ),
        )

    def test_extract_default_with_ast_default(self) -> None:
        """Tests that `extract_default` works when default parses to an AST type"""
        sample = (
            "maximal number of function evaluations per optimization\n"
            "        step (default: max_iter * 1.25)."
        )
        self.assertTupleEqual(
            extract_default(sample, emit_default_doc=True),
            (sample, "max_iter * 1.25"),
        )

    def test_set_default_doc_none(self) -> None:
        """Tests that `set_default_doc` does nop whence no doc in param"""
        name_param = "foo", {}
        self.assertTupleEqual(set_default_doc(name_param), name_param)


unittest_main()
